# Copyright © 2025 The GNOME Music Developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from gettext import gettext as _
from gi.repository import GObject, Gtk

from gnomemusic.coverpaintable import CoverPaintable
from gnomemusic.gstplayer import Playback
from gnomemusic.utils import ArtSize, DefaultIconType
from gnomemusic.player import Player
from gnomemusic.widgets.repeatmodebutton import RepeatModeButton  # noqa: F401
from gnomemusic.widgets.smoothscale import SmoothScale  # noqa: F401
from gnomemusic.widgets.twolinetip import TwoLineTip
from gnomemusic.widgets.volumebutton import VolumeButton  # noqa: F401
import gnomemusic.utils as utils


@Gtk.Template(resource_path='/org/gnome/Music/ui/PlayerToolbar.ui')
class PlayerToolbar(Gtk.ActionBar):
    """Main Player widget object

    Contains the ui of playing a song with Music.
    """

    __gtype_name__ = 'PlayerToolbar'

    _artist_label = Gtk.Template.Child()
    _cover_image = Gtk.Template.Child()
    _duration_label = Gtk.Template.Child()
    _next_button = Gtk.Template.Child()
    _play_button = Gtk.Template.Child()
    _play_pause_image = Gtk.Template.Child()
    _prev_button = Gtk.Template.Child()
    _progress_scale = Gtk.Template.Child()
    _progress_time_label = Gtk.Template.Child()
    _repeat_mode_button = Gtk.Template.Child()
    _song_info_box = Gtk.Template.Child()
    _title_label = Gtk.Template.Child()
    _volume_button = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        self._player = None

        self._cover_image.set_size_request(
            ArtSize.SMALL.width, ArtSize.SMALL.height)
        self._cover_image.props.pixel_size = ArtSize.SMALL.height
        self._cover_image.props.paintable = CoverPaintable(
            self, ArtSize.SMALL, DefaultIconType.ALBUM)

        self._tooltip = TwoLineTip()

    # FIXME: This is a workaround for not being able to pass the player
    # object via init when using Gtk.Builder.
    @GObject.Property(type=Player, default=None)
    def player(self):
        """The GstPlayer object used

        :return: player object
        :rtype: GstPlayer
        """
        return self._player

    @player.setter  # type: ignore
    def player(self, player):
        """Set the GstPlayer object used

        :param GstPlayer player: The GstPlayer to use
        """
        if (player is None
                or (self._player is not None
                    and self._player != player)):
            return

        self._player = player
        self._progress_scale.props.player = self._player

        self._player.connect('song-changed', self._update_view)
        self._player.connect(
            "notify::repeat-mode", self._on_repeat_mode_changed)
        self._player.connect('notify::state', self._sync_playing)
        self._repeat_mode_button.props.player = self._player

        self._player.bind_property(
            "volume", self._volume_button, "volume",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

        self._player.bind_property(
            "mute", self._volume_button, "mute",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

    @Gtk.Template.Callback()
    def _on_progress_value_changed(self, progress_scale):
        seconds = int(progress_scale.get_value() / 60)
        self._progress_time_label.set_label(utils.seconds_to_string(seconds))

    @Gtk.Template.Callback()
    def _on_prev_button_clicked(self, button):
        self._player.previous()

    @Gtk.Template.Callback()
    def _on_play_button_clicked(self, button):
        self._player.play_pause()

    @Gtk.Template.Callback()
    def _on_next_button_clicked(self, button):
        self._player.next()

    def _sync_playing(self, player, state):
        if (self._player.props.state == Playback.STOPPED
                and not self._player.props.has_next
                and not self._player.props.has_previous):
            self.props.revealed = False
            return

        self.props.revealed = True

        if self._player.props.state == Playback.PLAYING:
            icon_name = "media-playback-pause-symbolic"
            tooltip = _("Pause")
        else:
            icon_name = "media-playback-start-symbolic"
            tooltip = _("Play")

        if self._play_pause_image.props.icon_name != icon_name:
            self._play_pause_image.props.icon_name = icon_name

        self._play_button.set_tooltip_text(tooltip)

    def _sync_prev_next(self):
        self._next_button.props.sensitive = self._player.props.has_next
        self._prev_button.props.sensitive = self._player.props.has_previous

    def _update_view(self, player):
        """Update all visual elements on song change

        :param Player player: The main player object
        """
        coresong = player.props.current_song
        self._duration_label.props.label = utils.seconds_to_string(
            coresong.props.duration)
        self._progress_time_label.props.label = "0:00"

        self._play_button.set_sensitive(True)
        self._sync_prev_next()

        artist = coresong.props.artist
        title = coresong.props.title

        self._title_label.props.label = title
        self._artist_label.props.label = artist

        self._tooltip.props.title = title
        self._tooltip.props.subtitle = artist

        self._cover_image.props.paintable.props.coreobject = coresong

    @Gtk.Template.Callback()
    def _on_tooltip_query(self, widget, x, y, kb, tooltip, data=None):
        tooltip.set_custom(self._tooltip)

        return True

    def _on_repeat_mode_changed(
            self, player: Player, pspec: GObject.ParamSpec) -> None:
        self._sync_prev_next()



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Optional

from gi.repository import Gio, GLib, GObject, Gtk

from gnomemusic.player import Player, RepeatMode


@Gtk.Template(resource_path="/org/gnome/Music/ui/RepeatModeButton.ui")
class RepeatModeButton(Gtk.Box):
    """Repeat mode button widget."""

    __gtype_name__ = "RepeatModeButton"

    _menu_button = Gtk.Template.Child()

    def __init__(self) -> None:
        """Initialize the repeat mode button.
        """
        super().__init__()
        self._player: Optional[Player] = None

        menu = Gio.Menu.new()
        for mode in RepeatMode:
            item = Gio.MenuItem.new()
            item.set_label(mode.label)
            item.set_action_and_target_value(
                "repeatmenu.repeat", GLib.Variant("s", str(mode.value)))
            menu.append_item(item)

        self._menu_button.props.menu_model = menu

        self._repeat_action: Gio.SimpleAction = Gio.SimpleAction.new_stateful(
            "repeat", GLib.VariantType.new("s"), GLib.Variant("s", "0"))
        action_group = Gio.SimpleActionGroup()
        action_group.add_action(self._repeat_action)
        self.insert_action_group("repeatmenu", action_group)

        self._repeat_action.connect("activate", self._on_repeat_menu_changed)

    @GObject.Property(type=Player)
    def player(self) -> Optional[Player]:
        """Player object getter

        :returns: the Player object.
        :rtype: Player
        """
        return self._player

    @player.setter  # type: ignore
    def player(self, player: Player) -> None:
        """Player object setter

        :param Player player: the Player object
        """
        self._player = player

        self._player.bind_property(
            "repeat-mode", self._menu_button, "icon-name",
            GObject.BindingFlags.DEFAULT
            | GObject.BindingFlags.SYNC_CREATE,
            lambda binding, value: value.icon,
            None)

        repeat_mode = self._player.props.repeat_mode
        self._repeat_action.set_state(
            GLib.Variant("s", str(repeat_mode.value)))

        self._player.bind_property(
            "repeat-mode", self._repeat_action, "state",
            GObject.BindingFlags.DEFAULT
            | GObject.BindingFlags.SYNC_CREATE,
            lambda binding, value: GLib.Variant("s", str(value.value)),
            None)

    def _on_repeat_menu_changed(
            self, action: Gio.SimpleAction, new_state: GLib.Variant) -> None:
        if self._player is None:
            return

        self._repeat_action.set_state(new_state)
        new_mode = RepeatMode(int(new_state.get_string()))
        self._player.props.repeat_mode = new_mode



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Callable, List, Optional, Union
import typing
import asyncio

import gi
gi.require_versions({"Grl": "0.3", "Tsparql": "3.0"})
from gi.repository import Gio, Gtk, GLib, GObject, Tsparql

from gnomemusic.corealbum import CoreAlbum
from gnomemusic.coreartist import CoreArtist
from gnomemusic.coredisc import CoreDisc
from gnomemusic.coresong import CoreSong
from gnomemusic.grilowrappers.localsearchplaylists import LocalSearchPlaylists
from gnomemusic.trackerwrapper import TrackerWrapper
import gnomemusic.utils as utils
if typing.TYPE_CHECKING:
    from gi.repository import TSparql
    from gnomemusic.application import Application
    from gnomemusic.grilowrappers.playlist import Playlist
    CoreObject = Union[CoreAlbum, CoreArtist, CoreSong]


class LocalSearchWrapper(GObject.Object):
    """LocalSearch based data source
    """

    __gtype_name__ = "LocalSearchWrapper"

    _SPLICE_SIZE = 100

    def __init__(
            self, application: Application,
            tsparql_wrapper: TrackerWrapper) -> None:
        """Initialize LocalSearchWrapper

        Set up and fill models with information retrieved through
        LocalSearch.

       :param Application application: The Application instance
       :param TrackerWrapper tsparql_wrapper: The TrackerWrapper instance
        """
        super().__init__()

        self._application = application
        self._log = application.props.log
        self._notificationmanager = application.props.notificationmanager
        self._tsparql = tsparql_wrapper.props.local_db
        self._tsparql_playlists: Optional[LocalSearchPlaylists] = None
        self._tsparql_wrapper = tsparql_wrapper

        self._cancellable = Gio.Cancellable()

        self._albums_model = Gio.ListStore.new(CoreAlbum)
        self._artists_model = Gio.ListStore.new(CoreArtist)
        self._songs_model = Gio.ListStore.new(CoreSong)

        cm = application.props.coremodel
        cm.props.albums_proxy.append(self._albums_model)
        cm.props.artists_proxy.append(self._artists_model)
        cm.props.songs_proxy.append(self._songs_model)

        self._albums_search_model = Gtk.FilterListModel.new(self._albums_model)
        self._albums_search_model.set_filter(Gtk.AnyFilter())
        cm.props.albums_search_proxy.append(self._albums_search_model)

        self._artists_search_model = Gtk.FilterListModel.new(
            self._artists_model)
        self._artists_search_model.set_filter(Gtk.AnyFilter())
        cm.props.artists_search_proxy.append(self._artists_search_model)

        self._songs_search_model = Gtk.FilterListModel.new(self._songs_model)
        self._songs_search_model.set_filter(Gtk.AnyFilter())
        cm.props.songs_search_proxy.append(self._songs_search_model)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/albums.rq")
        self._albums_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/artists.rq")
        self._artists_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/songs.rq")
        prep_stmt = prep_stmt.replace("song_bind", "BIND(~song AS ?song)")
        self._song_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/songs.rq")
        prep_stmt = prep_stmt.replace("song_bind", "")
        self._songs_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/album_discs.rq")
        self._album_discs_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/album_disc.rq")
        self._album_disc_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/artist_albums.rq")
        self._artist_albums_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/search_albums.rq")
        self._search_albums_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/search_artists.rq")
        self._search_artists_stmt = self._tsparql.query_statement(prep_stmt)

        prep_stmt = self._prepare_statement(
            "/org/gnome/Music/queries/search_songs.rq")
        self._search_songs_stmt = self._tsparql.query_statement(prep_stmt)

        asyncio.create_task(self._init_albums_model())
        asyncio.create_task(self._init_artists_model())
        asyncio.create_task(self._init_songs_model())

        self._notifier = self._tsparql_wrapper.props.miner_fs.create_notifier()
        self._notifier.connect("events", self._on_notifier_event)

    def _prepare_statement(self, resource_path: str) -> str:
        """Helper to insert bus name and location filter in query"""
        gbytes = Gio.resources_lookup_data(
            resource_path, Gio.ResourceLookupFlags.NONE)
        query_str = gbytes.get_data().decode("utf-8")
        query_str = query_str.replace(
            "{bus_name}", self._tsparql_wrapper.props.miner_fs_busname)
        query_str = query_str.replace(
            "{location_filter}", self._tsparql_wrapper.location_filter())

        return query_str

    async def _init_albums_model(self) -> None:
        async with self._notificationmanager:
            try:
                cursor = await self._albums_stmt.execute_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                has_next = False
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                corealbum = CoreAlbum(self._application, cursor_dict)

                self._albums_model.append(corealbum)

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                    has_next = False

            cursor.close()

    async def _init_artists_model(self) -> None:
        async with self._notificationmanager:
            try:
                cursor = await self._artists_stmt.execute_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                has_next = False
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                coreartist = CoreArtist(self._application, cursor_dict)

                self._artists_model.append(coreartist)

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                    has_next = False

        cursor.close()

    async def _init_songs_model(self) -> None:
        async with self._notificationmanager:
            try:
                cursor = await self._songs_stmt.execute_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            songs: List[CoreSong] = []
            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                has_next = False
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                coresong = CoreSong(self._application, cursor_dict)

                songs.append(coresong)
                if len(songs) == self._SPLICE_SIZE:
                    self._songs_model.splice(
                        self._songs_model.get_n_items(), 0, songs)
                    songs.clear()

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                    has_next = False

            self._songs_model.splice(
                self._songs_model.get_n_items(), 0, songs)
            cursor.close()

            # Initialize the playlists subwrapper after the initial
            # songs model fill, the playlists expect a filled songs
            # model.
            self._tsparql_playlists = LocalSearchPlaylists(
                self._application, self._tsparql_wrapper, self._songs_model)

    def _find_equal_coreobject_urn(
            self, coreobj_compared: CoreObject,
            coreobj_provided: CoreObject) -> bool:
        return coreobj_compared.props.id == coreobj_provided.props.id

    async def _fileid_event(self, event: Tsparql.NotifierEvent) -> None:
        event_type = event.get_event_type()
        urn = event.get_urn()
        coresong_compare = CoreSong(self._application, {"id": urn})
        if event_type in [
                Tsparql.NotifierEventType.CREATE,
                Tsparql.NotifierEventType.UPDATE]:
            await self._add_song(urn)

            found, position = self._songs_model.find_with_equal_func(
                coresong_compare, self._find_equal_coreobject_urn)
            if found:
                coresong = self._songs_model.get_item(position)
                await self._update_album(coresong)
        elif event_type == Tsparql.NotifierEventType.DELETE:
            found, position = self._songs_model.find_with_equal_func(
                coresong_compare, self._find_equal_coreobject_urn)
            if found:
                coresong = self._songs_model.get_item(position)
                self._songs_model.remove(position)
                await self._update_album(coresong)

        if self._tsparql_playlists is not None:
            self._tsparql_playlists.check_smart_playlist_change()

    def _on_notifier_event(
            self, notifier: TSparql.Notifier, service: str, graph: str,
            events: set[Tsparql.NotifierEvent]) -> None:
        for event in events:
            urn = event.get_urn()
            if urn.startswith("urn:fileid:"):
                asyncio.create_task(self._fileid_event(event))

    async def _update_album(self, coresong: CoreSong) -> None:
        corealbum_compare = CoreAlbum(
            self._application, {"id": coresong.props.album_urn})
        found, position = self._albums_model.find_with_equal_func(
            corealbum_compare, self._find_equal_coreobject_urn)
        if found:
            self._albums_model[position].remove_song_from_album(
                coresong.props.album_disc_number, coresong.props.id)

    async def _add_song(self, urn: str) -> None:
        self._song_stmt.bind_string("song", urn)
        async with self._notificationmanager:
            try:
                cursor = await self._song_stmt.execute_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                has_next = False

            coresong_compare = CoreSong(self._application, {"id": urn})
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                coresong = CoreSong(self._application, cursor_dict)

                found, position = self._songs_model.find_with_equal_func(
                    coresong_compare, self._find_equal_coreobject_urn)
                if found:
                    self._songs_model.get_item(position).update(cursor_dict)
                else:
                    self._songs_model.append(coresong)

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                    has_next = False

            cursor.close()

    async def _get_album_discs(
            self, corealbum: CoreAlbum, disc_model: Gtk.SortListModel) -> None:
        async with self._notificationmanager:
            self._album_discs_stmt.bind_string("album_id", corealbum.props.id)
            try:
                cursor = await self._album_discs_stmt.execute_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                has_next = False
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                disc_nr = utils.get_int_from_cursor_dict(
                    cursor_dict, "albumDiscNumber")
                coredisc = CoreDisc(
                    self._application, corealbum, disc_nr)

                disc_model.append(coredisc)

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                    has_next = False

            cursor.close()

    def get_album_discs(
            self, corealbum: CoreAlbum, disc_model: Gtk.SortListModel) -> None:
        """Get all discs of an album

        :param CoreAlbum corealbum: The album
        :param Gtk.SortListModel disc_model: The model to fill
        """
        asyncio.create_task(self._get_album_discs(corealbum, disc_model))

    async def _get_album_disc(
            self, coredisc: CoreDisc, model: Gtk.FilterListModel) -> None:
        async with self._notificationmanager:
            self._album_disc_stmt.bind_string("album_id", coredisc.props.id)
            self._album_disc_stmt.bind_int("disc_nr", coredisc.props.disc_nr)
            try:
                cursor = await self._album_disc_stmt.execute_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            disc_song_ids = []
            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                has_next = False
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                disc_song_ids.append(cursor_dict.get("id"))

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                    has_next = False

            cursor.close()

            def _filter_func(coresong: CoreSong) -> bool:
                return coresong.props.id in disc_song_ids

            custom_filter = Gtk.CustomFilter()
            custom_filter.set_filter_func(_filter_func)
            model.set_filter(custom_filter)

    def get_album_disc(
            self, coredisc: CoreDisc, model: Gtk.FilterListModel) -> None:
        """Get all songs of an album disc

        :param CoreDisc coredisc: The album disc to look up
        :param Gtk.FilterListModel model: The model to fill
        """
        asyncio.create_task(self._get_album_disc(coredisc, model))

    async def _get_artist_albums(
            self, coreartist: CoreArtist, model: Gtk.FilterListModel) -> None:
        async with self._notificationmanager:
            self._artist_albums_stmt.bind_string("artist", coreartist.props.id)
            try:
                cursor = await self._artist_albums_stmt.execute_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            album_ids = []
            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                has_next = False
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                album_ids.append(cursor_dict.get("id"))

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                    has_next = False

            cursor.close()

        def albums_filter(corealbum: CoreAlbum, albums: List[str]) -> bool:
            return corealbum.props.id in albums

        custom_filter = Gtk.CustomFilter()
        custom_filter.set_filter_func(albums_filter, album_ids)
        model.set_filter(custom_filter)

    def get_artist_albums(
            self, coreartist: CoreArtist, model: Gtk.FilterListModel) -> None:
        """Get all albums by an artist

        :param CoreArtist coreartist: The artist to look up
        :param Gtk.FilterListModel model: The model to fill
        """
        asyncio.create_task(self._get_artist_albums(coreartist, model))

    def search(self, text: str) -> None:
        """Search for the given string in the wrappers

        If an empty string is provided, the wrapper should
        reset to an empty state.

        :param str text: The search string
        """
        self._cancellable.cancel()
        if text == "":
            self._artists_search_model.set_filter(Gtk.AnyFilter())
            self._albums_search_model.set_filter(Gtk.AnyFilter())
            self._songs_search_model.set_filter(Gtk.AnyFilter())
            return

        self._cancellable = Gio.Cancellable.new()
        asyncio.create_task(self._search_generic(
            text, self._cancellable, self._search_artists_stmt,
            self._artists_search_model))
        asyncio.create_task(self._search_generic(
            text, self._cancellable, self._search_albums_stmt,
            self._albums_search_model))
        asyncio.create_task(self._search_generic(
            text, self._cancellable, self._search_songs_stmt,
            self._songs_search_model))

    async def _search_generic(
            self, term: str, cancellable: Gio.Cancellable,
            statement: TSparql.SparqlStatement,
            model: Gtk.FilterListModel) -> None:
        """Search and fill the model with results"""
        async with self._notificationmanager:
            statement.bind_string("name", term)
            try:
                cursor = await statement.execute_async(cancellable)
            except GLib.Error as error:
                if not error.matches(
                        Gio.io_error_quark(), Gio.IOErrorEnum.CANCELLED):
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                return

            filter_ids = []
            try:
                has_next = await cursor.next_async(cancellable)
            except GLib.Error as error:
                if not error.matches(
                        Gio.io_error_quark(), Gio.IOErrorEnum.CANCELLED):
                    self._log.warning(
                        f"Error: {error.domain}, {error.message}")
                has_next = False
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                filter_ids.append(cursor_dict.get("id"))

                try:
                    has_next = await cursor.next_async(cancellable)
                except GLib.Error as error:
                    if not error.matches(
                            Gio.io_error_quark(), Gio.IOErrorEnum.CANCELLED):
                        self._log.warning(
                            f"Error: {error.domain}, {error.message}")
                    filter_ids = []
                    has_next = False

            cursor.close()

            def filter_func(coreobject: CoreObject) -> bool:
                return coreobject.props.id in filter_ids

            if len(filter_ids) == 0:
                custom_filter = Gtk.AnyFilter()
            else:
                custom_filter = Gtk.CustomFilter()
                custom_filter.set_filter_func(filter_func)

            model.set_filter(custom_filter)
            # If a search does not change the number of items found,
            # SearchView will not update without a signal.
            model.emit("items-changed", 0, 0, 0)

    def stage_playlist_deletion(self, playlist: Optional[Playlist]) -> None:
        """Prepares playlist deletion.

        :param Playlist playlist: playlist
        """
        if self._tsparql_playlists is None:
            return

        self._tsparql_playlists.stage_playlist_deletion(playlist)

    def finish_playlist_deletion(
            self, playlist: Playlist, deleted: bool) -> None:
        """Finishes playlist deletion.

        :param Playlist playlist: playlist
        :param bool deleted: indicates if the playlist has been deleted
        """
        if self._tsparql_playlists is None:
            return

        self._tsparql_playlists.finish_playlist_deletion(playlist, deleted)

    def create_playlist(
            self, playlist_title: str,
            callback: Callable[[Playlist], None]) -> None:
        """Creates a new user playlist.

        :param str playlist_title: playlist title
        :param callback: function to perform once, the playlist is created
        """
        if self._tsparql_playlists is None:
            return

        self._tsparql_playlists.create_playlist(playlist_title, callback)



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations

from gi.repository import GLib, GObject, GstAudio, Gtk


@Gtk.Template(resource_path="/org/gnome/Music/ui/VolumeButton.ui")
class VolumeButton(Gtk.Box):
    """Volume button widget with a mute button and a slider."""

    __gtype_name__ = "VolumeButton"

    _adjustment = Gtk.Template.Child()
    _menu_button = Gtk.Template.Child()
    _mute_button = Gtk.Template.Child()

    mute = GObject.Property(type=bool, default=False)
    volume = GObject.Property(type=float, default=1.0)

    def __init__(self) -> None:
        """Initialize the volume button.
        """
        super().__init__()

        self.bind_property(
            "mute", self._mute_button, "active",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

        self._adjustment_id = self._adjustment.connect(
            "value-changed", self._on_adjustment_changed)
        self._mute_id = self.connect(
            "notify::mute", self._on_mute_changed)
        self._volume_id = self.connect(
            "notify::volume", self._on_volume_changed)

    def _on_adjustment_changed(self, adjustment: Gtk.Adjustment) -> None:
        with GObject.signal_handler_block(self, self._volume_id):
            self.props.volume = self._cubic_to_linear(adjustment.props.value)

        if self.props.volume == 0:
            self.props.mute = True
        elif self.props.mute:
            self.props.mute = False

        GLib.idle_add(self._update_icon)

    def _on_mute_changed(
            self, widget: VolumeButton, pspec: GObject.ParamSpec) -> None:
        if (not self.props.mute
                and self.props.volume == 0):
            with GObject.signal_handler_block(self, self._volume_id):
                self.props.volume = self._cubic_to_linear(0.25)

        GLib.idle_add(self._set_adjustment)
        GLib.idle_add(self._update_icon)

    def _on_volume_changed(
            self, widget: VolumeButton, pspec: GObject.ParamSpec) -> None:
        if (not self.props.mute
                and self.props.volume == 0):
            with GObject.signal_handler_block(self, self._mute_id):
                self.props.mute = True

        GLib.idle_add(self._set_adjustment)

    def _cubic_to_linear(self, value: float) -> float:
        return GstAudio.stream_volume_convert_volume(
            GstAudio.StreamVolumeFormat.CUBIC,
            GstAudio.StreamVolumeFormat.LINEAR,
            value
        )

    def _linear_to_cubic(self, value: float) -> float:
        return GstAudio.stream_volume_convert_volume(
            GstAudio.StreamVolumeFormat.LINEAR,
            GstAudio.StreamVolumeFormat.CUBIC,
            value
        )

    def _set_adjustment(self) -> bool:
        volume = self.props.volume
        if self.props.mute:
            value = 0.
        else:
            value = self._linear_to_cubic(volume)

        with GObject.signal_handler_block(
                self._adjustment, self._adjustment_id):
            self._adjustment.props.value = value

        return GLib.SOURCE_REMOVE

    def _update_icon(self) -> None:
        volume_cubic = self._linear_to_cubic(self.props.volume)

        if self.props.mute:
            icon_name = "audio-volume-muted-symbolic"
        elif volume_cubic < 0.3:
            icon_name = "audio-volume-low-symbolic"
        elif volume_cubic < 0.7:
            icon_name = "audio-volume-medium-symbolic"
        else:
            icon_name = "audio-volume-high-symbolic"

        self._menu_button.props.icon_name = icon_name



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Any, Dict, Optional
import typing

from gi.repository import GLib, GObject

from gnomemusic.songart import SongArt
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application
    from gnomemusic.coregrilo import CoreGrilo
import gnomemusic.utils as utils


class CoreSong(GObject.GObject):
    """Song information object

    Contains all relevant information of a song.
    """

    __gtype_name__ = "CoreSong"

    class Validation(GObject.GEnum):
        """Song validation status"""
        PENDING = 0
        IN_PROGRESS = 1
        FAILED = 2
        SUCCEEDED = 3

    album = GObject.Property(type=str)
    album_urn = GObject.Property(type=str)
    album_disc_number = GObject.Property(type=int)
    artist = GObject.Property(type=str)
    cursor_dict = GObject.Property()
    duration = GObject.Property(type=int)
    id = GObject.Property(type=str, default=None)
    play_count = GObject.Property(type=int)
    state = GObject.Property()  # FIXME: How to set an IntEnum type?
    title = GObject.Property(type=str)
    track_number = GObject.Property(type=int)
    url = GObject.Property(type=str)
    validation = GObject.Property(type=Validation, default=Validation.PENDING)

    def __init__(
            self, application: Application,
            cursor_dict: Dict[str, Any]) -> None:
        """Initiate the CoreSong object

        :param Application application: The application object
        :param Dict[str, Any] cursor_dict: Dict with Tsparql keys
        """
        super().__init__()

        self._application: Application = application
        self._coregrilo: CoreGrilo = application.props.coregrilo
        self._favorite: bool = False
        self._last_played: Optional[GLib.DateTime] = None
        self._thumbnail: Optional[str] = None

        self.props.id = cursor_dict.get("id")
        self.update(cursor_dict)

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, CoreSong)
                and other.props.id == self.props.id)

    @GObject.Property(type=bool, default=False)
    def favorite(self) -> bool:
        return self._favorite

    @favorite.setter  # type: ignore
    def favorite(self, favorite: bool) -> None:
        if self._favorite == favorite:
            return

        self._favorite = favorite
        self._coregrilo.writeback_tracker(self, "favorite")

    @GObject.Property(type=GLib.DateTime, default=None)
    def last_played(self) -> Optional[GLib.DateTime]:
        """Get last played time

        :returns: Last played date time if available
        :rtype: GLib.DateTime or None
        """
        return self._last_played

    @last_played.setter  # type: ignore
    def last_played(self, value: Optional[GLib.DateTime]) -> None:
        """Set last played time

        :param GLib.DateTime value: The datetime to set
        """
        if not value:
            return

        self._last_played = value
        self._coregrilo.writeback_tracker(self, "last-played")

    @GObject.Property(type=str, default=None)
    def thumbnail(self) -> str:
        """Song art thumbnail retrieval

        :return: The song art uri or "generic"
        :rtype: string
        """
        if self._thumbnail is None:
            self._thumbnail = "generic"
            SongArt(self._application, self)

        return self._thumbnail

    @thumbnail.setter  # type: ignore
    def thumbnail(self, value: str) -> None:
        """Song art thumbnail setter

        :param string value: uri or "generic"
        """
        self._thumbnail = value

    def update(self, cursor_dict: Dict[str, Any]) -> None:
        """Update the song with information from the dictionary

        :param Dict[str, Any] cursor_dict: The dicationary to use
        """
        self.props.album = cursor_dict.get("album") or ""
        self.props.album_urn = cursor_dict.get("album_urn")
        self.props.album_disc_number = utils.get_int_from_cursor_dict(
            cursor_dict, "albumDiscNumber")
        self.props.artist = utils.get_artist_from_cursor_dict(cursor_dict)
        self.props.duration = utils.get_int_from_cursor_dict(
            cursor_dict, "duration")
        self._favorite = bool(cursor_dict.get("favorite"))
        self._last_played = cursor_dict.get("lastPlayed")
        self.props.play_count = utils.get_int_from_cursor_dict(
            cursor_dict, "playCount")
        self.props.title = utils.get_title_from_cursor_dict(cursor_dict)
        self.props.track_number = utils.get_int_from_cursor_dict(
            cursor_dict, "trackNumber")
        self.props.url = cursor_dict.get("url")

        self.props.cursor_dict = cursor_dict

    def bump_play_count(self) -> None:
        self.props.play_count = self.props.play_count + 1
        self._coregrilo.writeback_tracker(self, "play-count")



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from enum import IntEnum
from random import randint
from typing import Dict, Optional
import typing

from gi.repository import GObject, GstPbutils

from gnomemusic.coresong import CoreSong
from gnomemusic.utils import RepeatMode
from gnomemusic.widgets.songwidget import SongWidget

if typing.TYPE_CHECKING:
    from gi.repository import GLib

    from gnomemusic.application import Application
    from gnomemusic.coremodel import CoreModel


class Queue(GObject.GObject):
    """Queue of songs to be played

    Contains the logic to validate a song, handle RepeatMode and the
    list of songs being played.
    """

    class Type(IntEnum):
        """Type of queue"""
        SONGS = 0
        ALBUM = 1
        ARTIST = 2
        PLAYLIST = 3
        SEARCH_RESULT = 4

    repeat_mode = GObject.Property(type=object)

    def __init__(self, application: Application) -> None:
        """Initialize the queue
        """
        super().__init__()

        GstPbutils.pb_utils_init()

        self._app = application
        self._log = application.props.log
        self._position = 0

        self._validation_songs: Dict[str, CoreSong] = {}
        self._discoverer = GstPbutils.Discoverer()
        self._discoverer.connect("discovered", self._on_discovered)
        self._discoverer.start()

        self._coremodel = self._app.props.coremodel
        self._model = self._coremodel.props.queue
        self._model_recent = self._coremodel.props.recent_queue

        self._coremodel.connect("queue-loaded", self._on_queue_loaded)

        self.connect("notify::repeat-mode", self._on_repeat_mode_changed)

    def has_next(self) -> bool:
        """Check if there is a song after the current one.

        :return: True if there is a song. False otherwise.
        :rtype: bool
        """
        if (self.props.repeat_mode == RepeatMode.SONG
                or self.props.repeat_mode == RepeatMode.ALL
                or self.props.position < self._model.get_n_items() - 1):
            return True

        return False

    def has_previous(self) -> bool:
        """Check if there is a song before the current one.

        :return: True if there is a song. False otherwise.
        :rtype: bool
        """
        if (self.props.repeat_mode == RepeatMode.SONG
                or self.props.repeat_mode == RepeatMode.ALL
                or (self.props.position <= self._model.get_n_items() - 1
                    and self.props.position > 0)):
            return True

        return False

    def get_next(self) -> Optional[CoreSong]:
        """Get the next song in the queue.

        :return: The next CoreSong or None.
        :rtype: CoreSong | None
        """
        if not self.has_next():
            return None

        if self.props.repeat_mode == RepeatMode.SONG:
            next_position = self.props.position
        elif (self.props.repeat_mode == RepeatMode.ALL
                and self.props.position == self._model.get_n_items() - 1):
            next_position = 0
        else:
            next_position = self.props.position + 1

        return self._model[next_position]

    def next(self) -> bool:
        """Go to the next song in the queue.

        :return: True if the operation succeeded. False otherwise.
        :rtype: bool
        """
        if not self.has_next():
            return False

        if self.props.repeat_mode == RepeatMode.SONG:
            next_position = self.props.position
        elif (self.props.repeat_mode == RepeatMode.ALL
                and self.props.position == self._model.get_n_items() - 1):
            next_position = 0
        else:
            next_position = self.props.position + 1

        self._model[self.props.position].props.state = SongWidget.State.PLAYED
        self._position = next_position

        next_song = self._model[next_position]
        if next_song.props.validation == CoreSong.Validation.FAILED:
            return self.next()

        self._update_model_recent()
        next_song.props.state = SongWidget.State.PLAYING
        self._validate_next_song()
        return True

    def previous(self) -> bool:
        """Go to the previous song in the queue.

        :return: True if the operation succeeded. False otherwise.
        :rtype: bool
        """
        if not self.has_previous():
            return False

        if self.props.repeat_mode == RepeatMode.SONG:
            previous_position = self.props.position
        elif (self.props.repeat_mode == RepeatMode.ALL
                and self.props.position == 0):
            previous_position = self._model.get_n_items() - 1
        else:
            previous_position = self.props.position - 1

        self._model[self.props.position].props.state = SongWidget.State.PLAYED
        self._position = previous_position

        previous_song = self._model[previous_position]
        if previous_song.props.validation == CoreSong.Validation.FAILED:
            return self.previous()

        self._update_model_recent()
        self._model[previous_position].props.state = SongWidget.State.PLAYING
        self._validate_previous_song()
        return True

    @GObject.Property(type=int, default=0, flags=GObject.ParamFlags.READABLE)
    def position(self) -> int:
        """Gets current song index.

        :returns: position of the current song in the queue.
        :rtype: int
        """
        return self._position

    @GObject.Property(
        type=CoreSong, default=None, flags=GObject.ParamFlags.READABLE)
    def current_song(self) -> Optional[CoreSong]:
        """Get current song.

        :returns: the song being played or None if there are no songs
        :rtype: CoreSong
        """
        n_items = self._model.get_n_items()
        if (n_items != 0
                and n_items > self._position):
            current_song = self._model[self._position]
            if current_song.props.state == SongWidget.State.PLAYING:
                return current_song

        for idx, coresong in enumerate(self._model):
            if coresong.props.state == SongWidget.State.PLAYING:
                self._position = idx
                self._update_model_recent()
                return coresong

        return None

    def set_song(self, song: Optional[CoreSong]) -> Optional[CoreSong]:
        """Sets current song.

        If no song is provided, a song is automatically selected.

        :param CoreSong song: song to set
        :returns: The selected song
        :rtype: CoreSong | None
        """
        n_items = self._model.get_n_items()
        if n_items == 0:
            return None

        if song is None:
            if self.props.repeat_mode == RepeatMode.SHUFFLE:
                self._model.shuffle(0, randint(0, n_items - 1))
            else:
                self._model.deshuffle()
            position = 0
            song = self._model.get_item(position)
            song.props.state = SongWidget.State.PLAYING
            self._position = position
            self._validate_song(song)
            self._validate_next_song()
            self._update_model_recent()
            return song

        self._model.deshuffle()
        for idx, coresong in enumerate(self._model):
            if coresong == song:
                coresong.props.state = SongWidget.State.PLAYING
                if self.props.repeat_mode == RepeatMode.SHUFFLE:
                    self._model.shuffle(0, idx)
                    self._position = 0
                else:
                    self._position = idx
                self._validate_song(song)
                self._validate_next_song()
                self._update_model_recent()
                return song

        return None

    def end(self) -> None:
        """End play of this queue
        """
        self._reset_queue_state()

    def _update_model_recent(self) -> None:
        recent_size = self._coremodel.props.recent_queue_size
        offset = max(0, self._position - recent_size)
        self._model_recent.set_offset(offset)

    def _on_repeat_mode_changed(
            self, queue: Queue, param: GObject.ParamSpecBoxed) -> None:
        if self.props.repeat_mode == RepeatMode.SHUFFLE:
            self._model.shuffle(self.props.position)
        elif self.props.repeat_mode in [RepeatMode.NONE, RepeatMode.ALL]:
            self._model.deshuffle(self.props.position)

    def _on_queue_loaded(
            self, coremodel: CoreModel, queue_type: Queue.Type) -> None:
        self._reset_queue_state()

    def _reset_queue_state(self) -> None:
        for song in self._model:
            song.props.state = SongWidget.State.UNPLAYED

    def _validate_song(self, coresong: CoreSong) -> None:
        # Song is being processed or has already been processed.
        # Nothing to do.
        if coresong.props.validation > CoreSong.Validation.PENDING:
            return

        url = coresong.props.url
        if not url:
            self._log.warning(
                "The item {} doesn't have a URL set.".format(coresong))
            return
        if not url.startswith("file://"):
            self._log.debug(
                "Skipping validation of {} as not a local file".format(url))
            return

        coresong.props.validation = CoreSong.Validation.IN_PROGRESS
        self._validation_songs[url] = coresong
        self._discoverer.discover_uri_async(url)

    def _validate_next_song(self) -> None:
        if self.props.repeat_mode == RepeatMode.SONG:
            return

        current_position = self.props.position
        next_position = current_position + 1
        if next_position == self._model.get_n_items():
            if self.props.repeat_mode != RepeatMode.ALL:
                return
            next_position = 0

        self._validate_song(self._model[next_position])

    def _validate_previous_song(self) -> None:
        if self.props.repeat_mode == RepeatMode.SONG:
            return

        current_position = self.props.position
        previous_position = current_position - 1
        if previous_position < 0:
            if self.props.repeat_mode != RepeatMode.ALL:
                return
            previous_position = self._model.get_n_items() - 1

        self._validate_song(self._model[previous_position])

    def _on_discovered(
            self, discoverer: GstPbutils.Discoverer,
            info: GstPbutils.DiscovererInfo, error: GLib.Error) -> None:
        url = info.get_uri()
        coresong = self._validation_songs[url]

        if error:
            self._log.warning("Info {}: error: {}".format(info, error))
            coresong.props.validation = CoreSong.Validation.FAILED
        else:
            coresong.props.validation = CoreSong.Validation.SUCCEEDED



# Copyright 2024 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Optional
import asyncio
import time

from gettext import gettext as _

from gi.repository import GLib, GObject, Gio

from gnomemusic.coresong import CoreSong
from gnomemusic.grilowrappers.playlist import Playlist
import gnomemusic.utils as utils


class SmartPlaylist(Playlist):
    """Base class for smart playlists"""

    __gtype_name__ = "SmartPlaylist"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.is_smart = True
        self._model: Optional[Gio.ListStore] = None

    @GObject.Property(type=Gio.ListStore, default=None)
    def model(self):
        if self._model is None:
            self._model = Gio.ListStore.new(CoreSong)
            asyncio.create_task(self._populate_model())

        return self._model

    async def _populate_model(self) -> None:
        async with self._notificationmanager:
            if self._model is None:
                return

            try:
                cursor = self._tsparql.query(self.props.query)
            except GLib.Error as error:
                self._log.warning(
                    f"Failure populating playlist model {self.props.pl_id}:"
                    f" {error.domain}, {error.message}")
                return

            try:
                has_next = await cursor.next_async()
            except GLib.Error as error:
                cursor.close()
                self._log.warning(
                    f"Cursor iteration error: {error.domain}, {error.message}")
                return
            while has_next:
                cursor_dict = utils.dict_from_cursor(cursor)
                coresong = CoreSong(self._application, cursor_dict)
                self._bind_to_main_song(coresong)
                if coresong not in self._songs_todelete:
                    self._model.append(coresong)

                try:
                    has_next = await cursor.next_async()
                except GLib.Error as error:
                    cursor.close()
                    self._log.warning(
                        f"Cursor iteration error: {error.domain}, "
                        f"{error.message}")
                    return
            else:
                cursor.close()
                self.props.count = self._model.get_n_items()

    def update(self):
        """Updates playlist model."""
        if self._model is None:
            return

        self._model.remove_all()
        asyncio.create_task(self._populate_model())

        return


class MostPlayed(SmartPlaylist):
    """Most Played smart playlist"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.tag_text = "MOST_PLAYED"
        # TRANSLATORS: this is a playlist name
        self._title = _("Most Played")
        self.props.icon_name = "audio-speakers-symbolic"
        self.props.query = """
        SELECT
            ?song AS ?id
            ?title
            ?url
            ?artist
            ?album
            ?duration
            ?trackNumber
            ?albumDiscNumber
            ?playCount
            ?tag AS ?favorite
        WHERE {
            SERVICE <dbus:%(miner_fs_busname)s> {
                GRAPH tracker:Audio {
                    SELECT
                        ?song
                        nie:title(?song) AS ?title
                        nie:isStoredAs(?song) AS ?url
                        nmm:artistName(nmm:artist(?song)) AS ?artist
                        nie:title(nmm:musicAlbum(?song)) AS ?album
                        nfo:duration(?song) AS ?duration
                        nmm:trackNumber(?song) AS ?trackNumber
                        nmm:setNumber(nmm:musicAlbumDisc(?song))
                            AS ?albumDiscNumber
                    WHERE {
                        ?song a nmm:MusicPiece .
                        %(location_filter)s
                    }
                }
            }
            ?song nie:usageCounter ?playCount
            OPTIONAL { ?song nao:hasTag ?tag .
                       FILTER (?tag = nao:predefined-tag-favorite) }
        }
        ORDER BY DESC(?playCount) LIMIT 50
        """.replace('\n', ' ').strip() % {
            "location_filter": self._tsparql_wrapper.location_filter(),
            "miner_fs_busname": self._tsparql_wrapper.props.miner_fs_busname,
        }


class NeverPlayed(SmartPlaylist):
    """Never Played smart playlist"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.tag_text = "NEVER_PLAYED"
        # TRANSLATORS: this is a playlist name
        self._title = _("Never Played")
        self.props.icon_name = "deaf-symbolic"
        self.props.query = """
        SELECT
            ?song AS ?id
            ?title
            ?url
            ?artist
            ?album
            ?duration
            ?trackNumber
            ?albumDiscNumber
            nie:usageCounter(?song) AS ?playCount
            ?tag AS ?favorite
        WHERE {
            SERVICE <dbus:%(miner_fs_busname)s> {
                GRAPH tracker:Audio {
                    SELECT
                        ?song
                        nie:title(?song) AS ?title
                        nie:isStoredAs(?song) AS ?url
                        nmm:artistName(nmm:artist(?song)) AS ?artist
                        nie:title(nmm:musicAlbum(?song)) AS ?album
                        nfo:duration(?song) AS ?duration
                        nmm:trackNumber(?song) AS ?trackNumber
                        nmm:setNumber(nmm:musicAlbumDisc(?song))
                            AS ?albumDiscNumber
                    WHERE {
                        ?song a nmm:MusicPiece .
                        %(location_filter)s
                    }
                }
            }
            FILTER ( NOT EXISTS { ?song nie:usageCounter ?count .} )
            OPTIONAL { ?song nao:hasTag ?tag .
                       FILTER (?tag = nao:predefined-tag-favorite) }
        } ORDER BY nfo:fileLastAccessed(?song) LIMIT 50
        """.replace('\n', ' ').strip() % {
            "location_filter": self._tsparql_wrapper.location_filter(),
            "miner_fs_busname": self._tsparql_wrapper.props.miner_fs_busname,
        }


class RecentlyPlayed(SmartPlaylist):
    """Recently Played smart playlist"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.tag_text = "RECENTLY_PLAYED"
        # TRANSLATORS: this is a playlist name
        self._title = _("Recently Played")
        self.props.icon_name = "document-open-recent-symbolic"

        sparql_midnight_dateTime_format = "%Y-%m-%dT00:00:00Z"
        days_difference = 7
        seconds_difference = days_difference * 86400
        compare_date = time.strftime(
            sparql_midnight_dateTime_format,
            time.gmtime(time.time() - seconds_difference))
        self.props.query = """
        SELECT
            ?song AS ?id
            ?title
            ?url
            ?artist
            ?album
            ?duration
            ?trackNumber
            ?albumDiscNumber
            nie:usageCounter(?song) AS ?playCount
            ?tag AS ?favorite
        WHERE {
            {
                SELECT
                    ?song
                    ?title
                    ?url
                    ?artist
                    ?album
                    ?duration
                    ?trackNumber
                    ?albumDiscNumber
                    ?playCount
                    ?tag
                    ?lastPlayed
                WHERE {
                    SERVICE <dbus:%(miner_fs_busname)s> {
                        GRAPH tracker:Audio {
                            SELECT
                                ?song
                                nie:title(?song) AS ?title
                                nie:isStoredAs(?song) AS ?url
                                nmm:artistName(nmm:artist(?song)) AS ?artist
                                nie:title(nmm:musicAlbum(?song)) AS ?album
                                nfo:duration(?song) AS ?duration
                                nmm:trackNumber(?song) AS ?trackNumber
                                nmm:setNumber(nmm:musicAlbumDisc(?song))
                                    AS ?albumDiscNumber
                            WHERE {
                                ?song a nmm:MusicPiece .
                                %(location_filter)s
                            }
                        }
                    }
                    ?song nie:contentAccessed ?lastPlayed ;
                        nie:usageCounter ?playCount .
                    OPTIONAL { ?song nao:hasTag ?tag .
                               FILTER (?tag = nao:predefined-tag-favorite) }
                } ORDER BY DESC(?lastPlayed) LIMIT 50
            }
            FILTER (?lastPlayed > '%(compare_date)s'^^xsd:dateTime)
        }
        """.replace('\n', ' ').strip() % {
            'compare_date': compare_date,
            "location_filter": self._tsparql_wrapper.location_filter(),
            "miner_fs_busname": self._tsparql_wrapper.props.miner_fs_busname,
        }


class RecentlyAdded(SmartPlaylist):
    """Recently Added smart playlist"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.tag_text = "RECENTLY_ADDED"
        # TRANSLATORS: this is a playlist name
        self._title = _("Recently Added")
        self.props.icon_name = "list-add-symbolic"

        sparql_midnight_dateTime_format = "%Y-%m-%dT00:00:00Z"
        days_difference = 7
        seconds_difference = days_difference * 86400
        compare_date = time.strftime(
            sparql_midnight_dateTime_format,
            time.gmtime(time.time() - seconds_difference))
        self.props.query = """
        SELECT
            ?song AS ?id
            ?title
            ?url
            ?artist
            ?album
            ?duration
            ?trackNumber
            ?albumDiscNumber
            nie:usageCounter(?song) AS ?playCount
            ?tag AS ?favorite
        WHERE {
            SERVICE <dbus:%(miner_fs_busname)s> {
                GRAPH tracker:Audio {
                    SELECT
                        ?song
                        nie:title(?song) AS ?title
                        nie:isStoredAs(?song) AS ?url
                        nmm:artistName(nmm:artist(?song)) AS ?artist
                        nie:title(nmm:musicAlbum(?song)) AS ?album
                        nfo:duration(?song) AS ?duration
                        nmm:trackNumber(?song) AS ?trackNumber
                        nmm:setNumber(nmm:musicAlbumDisc(?song))
                            AS ?albumDiscNumber
                        ?added
                    WHERE {
                        ?song a nmm:MusicPiece ;
                              nrl:added ?added .
                        %(location_filter)s
                        FILTER ( ?added > '%(compare_date)s'^^xsd:dateTime )
                    }
                }
            }
            OPTIONAL { ?song nao:hasTag ?tag .
                       FILTER (?tag = nao:predefined-tag-favorite) }
        } ORDER BY DESC(nrl:added(?song)) LIMIT 50
        """.replace('\n', ' ').strip() % {
            'compare_date': compare_date,
            "location_filter": self._tsparql_wrapper.location_filter(),
            "miner_fs_busname": self._tsparql_wrapper.props.miner_fs_busname,
        }


class Favorites(SmartPlaylist):
    """Favorites smart playlist"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.tag_text = "FAVORITES"
        # TRANSLATORS: this is a playlist name
        self._title = _("Starred Songs")
        self.props.icon_name = "starred-symbolic"
        self.props.query = """
            SELECT
                ?song AS ?id
                ?title
                ?url
                ?artist
                ?album
                ?duration
                ?trackNumber
                ?albumDiscNumber
                nie:usageCounter(?song) AS ?playCount
                nao:predefined-tag-favorite AS ?favorite
            WHERE {
                SERVICE <dbus:%(miner_fs_busname)s> {
                    GRAPH tracker:Audio {
                        SELECT
                            ?song
                            nie:title(?song) AS ?title
                            nie:isStoredAs(?song) AS ?url
                            nmm:artistName(nmm:artist(?song)) AS ?artist
                            nie:title(nmm:musicAlbum(?song)) AS ?album
                            nfo:duration(?song) AS ?duration
                            nmm:trackNumber(?song) AS ?trackNumber
                            nmm:setNumber(nmm:musicAlbumDisc(?song))
                                AS ?albumDiscNumber
                            nrl:added(?song) AS ?added
                        WHERE {
                            ?song a nmm:MusicPiece .
                            %(location_filter)s
                        }
                    }
                }
                ?song nao:hasTag nao:predefined-tag-favorite .
            } ORDER BY DESC(?added)
        """.replace('\n', ' ').strip() % {
            "location_filter": self._tsparql_wrapper.location_filter(),
            "miner_fs_busname": self._tsparql_wrapper.props.miner_fs_busname,
        }


class InsufficientTagged(SmartPlaylist):
    """Lacking tags to be displayed in the artist/album views"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.tag_text = "INSUFFICIENT_TAGGED"
        # TRANSLATORS: this is a playlist name indicating that the
        # files are not tagged enough to be displayed in the albums
        # or artists views.
        self._title = _("Insufficiently Tagged")
        self.props.icon_name = "question-round-symbolic"
        self.props.query = """
        SELECT
            ?song AS ?id
            ?title
            ?url
            ?artist
            ?album
            ?duration
            ?trackNumber
            ?albumDiscNumber
            nie:usageCounter(?song) AS ?playCount
            ?tag AS ?favorite
        WHERE {
            SERVICE <dbus:%(miner_fs_busname)s> {
                GRAPH tracker:Audio {
                    SELECT
                        ?song
                        nie:title(?song) AS ?title
                        nie:isStoredAs(?song) AS ?url
                        nmm:artistName(nmm:artist(?song)) AS ?artist
                        nie:title(nmm:musicAlbum(?song)) AS ?album
                        nfo:duration(?song) AS ?duration
                        nmm:trackNumber(?song) AS ?trackNumber
                        nmm:setNumber(nmm:musicAlbumDisc(?song))
                            AS ?albumDiscNumber
                    WHERE {
                        {
                            ?song a nmm:MusicPiece .
                            %(location_filter)s
                            FILTER NOT EXISTS {
                                ?song nmm:musicAlbum ?album
                            }
                        } UNION {
                            ?song a nmm:MusicPiece .
                            %(location_filter)s
                            FILTER NOT EXISTS {
                                ?song nmm:artist ?artist
                            }
                        }
                    }
                }
            }
            OPTIONAL { ?song nao:hasTag ?tag .
                       FILTER (?tag = nao:predefined-tag-favorite) }
        }
        """.replace('\n', ' ').strip() % {
            "location_filter": self._tsparql_wrapper.location_filter(),
            "miner_fs_busname": self._tsparql_wrapper.props.miner_fs_busname,
        }


class AllSongs(SmartPlaylist):
    """All Songs smart playlist"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.props.tag_text = "ALL_SONGS"
        # TRANSLATORS: this is a playlist name
        self._title = _("All Songs")
        self.props.icon_name = "folder-music-symbolic"

        self.props.query = """
        SELECT
            ?song AS ?id
            ?title
            ?url
            ?artist
            ?album
            ?duration
            ?trackNumber
            ?albumDiscNumber
            nie:usageCounter(?song) AS ?playCount
            ?tag AS ?favorite
        WHERE {
            SERVICE <dbus:%(miner_fs_busname)s> {
                GRAPH tracker:Audio {
                    SELECT
                        ?song
                        nie:title(?song) AS ?title
                        nie:isStoredAs(?song) AS ?url
                        nmm:artistName(nmm:artist(?song)) AS ?artist
                        nie:title(nmm:musicAlbum(?song)) AS ?album
                        nfo:duration(?song) AS ?duration
                        nmm:trackNumber(?song) AS ?trackNumber
                        nmm:setNumber(nmm:musicAlbumDisc(?song))
                            AS ?albumDiscNumber
                    WHERE {
                        ?song a nmm:MusicPiece .
                        %(location_filter)s
                    }
                }
            }
            OPTIONAL { ?song nao:hasTag ?tag .
                       FILTER (?tag = nao:predefined-tag-favorite) }
        } ORDER BY ?artist ?album ?trackNumber
        """.replace('\n', ' ').strip() % {
            "location_filter": self._tsparql_wrapper.location_filter(),
            "miner_fs_busname": self._tsparql_wrapper.props.miner_fs_busname,
        }



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Any, Dict
import typing

from gi.repository import GObject, Gio, Gtk

from gnomemusic.albumart import AlbumArt
from gnomemusic.coredisc import CoreDisc
import gnomemusic.utils as utils
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application


class CoreAlbum(GObject.GObject):
    """Album information object

    Contains all relevant information about an album.
    """

    __gtype_name__ = "CoreAlbum"

    artist = GObject.Property(type=str)
    composer = GObject.Property(type=str, default=None)
    duration = GObject.Property(type=int, default=0)
    id = GObject.Property(type=str)
    title = GObject.Property(type=str)
    url = GObject.Property(type=str)
    year = GObject.Property(type=str, default=None)

    def __init__(
            self, application: Application,
            cursor_dict: Dict[str, Any]) -> None:
        """Initiate the CoreAlbum object

        :param Application application: The application object
        :param Dict[str, Any] cursor_dict: Dict with Tsparql keys
        """
        super().__init__()

        self._application = application
        self._coregrilo = application.props.coregrilo
        self._model = None
        self._thumbnail = None

        self.update(cursor_dict)

    def update(self, cursor_dict):
        """Update the CoreAlbum object with new info

        :param Dict[str, Any] cursor_dict: Dict with Tsparql keys
        """
        self.props.artist = utils.get_artist_from_cursor_dict(cursor_dict)
        self.props.composer = cursor_dict.get("composer")
        self.props.id = cursor_dict.get("id")
        self.props.title = utils.get_title_from_cursor_dict(cursor_dict)
        self.props.url = cursor_dict.get("url")
        self.props.year = cursor_dict.get("publicationDate")

    def remove_song_from_album(self, disc_nr: int, song_id: str) -> None:
        """Removes given song on given album disc

        :param int disc_nr: Number of disc
        :param int song_id: Song identifier
        """
        for coredisc in self.props.model:
            if coredisc.props.disc_nr == disc_nr:
                coredisc.remove_song_from_disc(song_id)
                break

    def _get_album_model(self):
        disc_model = Gio.ListStore()
        disc_no_exp = Gtk.PropertyExpression.new(CoreDisc, None, "disc_nr")
        disc_sorter = Gtk.NumericSorter.new(disc_no_exp)
        disc_model_sort = Gtk.SortListModel.new(disc_model, disc_sorter)

        self._coregrilo.get_album_discs(self, disc_model)

        return disc_model_sort

    @GObject.Property(
        type=Gtk.SortListModel, default=None,
        flags=GObject.ParamFlags.READABLE)
    def model(self):
        if self._model is None:
            self._model = self._get_album_model()
            self._model.connect("items-changed", self._on_list_items_changed)
            self._model.items_changed(0, 0, self._model.get_n_items())

        return self._model

    def _on_list_items_changed(self, model, position, removed, added):
        with self.freeze_notify():
            if added > 0:
                for i in range(added):
                    coredisc = model[position + i]
                    coredisc.connect(
                        "notify::duration", self._on_duration_changed)

    def _on_duration_changed(self, coredisc, duration):
        duration = 0

        for coredisc in self.props.model:
            duration += coredisc.props.duration

        self.props.duration = duration

    @GObject.Property(type=str, default=None)
    def thumbnail(self):
        """Album art thumbnail retrieval

        :return: The album art uri or "generic"
        :rtype: string
        """
        if self._thumbnail is None:
            self._thumbnail = "generic"
            AlbumArt(self._application, self)

        return self._thumbnail

    @thumbnail.setter  # type: ignore
    def thumbnail(self, value):
        """Album art thumbnail setter

        :param string value: uri or "generic"
        """
        self._thumbnail = value

    @GObject.Property(type=object, flags=GObject.ParamFlags.READABLE)
    def corealbum(self) -> CoreAlbum:
        return self



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
import typing

from gi.repository import GObject, Gtk

from gnomemusic.coresong import CoreSong
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application
    from gnomemusic.corealbum import CoreAlbum


class CoreDisc(GObject.GObject):

    __gtype_name__ = "CoreDisc"

    id = GObject.Property(type=str)
    disc_nr = GObject.Property(type=int, default=0)
    duration = GObject.Property(type=int, default=None)

    def __init__(
            self, application: Application, corealbum: CoreAlbum,
            nr: int) -> None:
        """Initialize a CoreDisc object

        :param Application application: The application object
        :param CoreAlbum corealbum: The album to create a disc from
        :param int nr: The disc number to create an object for
        """
        super().__init__()

        self._coregrilo = application.props.coregrilo
        self._coremodel = application.props.coremodel
        self._log = application.props.log
        self._model = None

        self.update(corealbum)
        self.props.disc_nr = nr

    def update(self, corealbum: CoreAlbum) -> None:
        self.props.id = corealbum.props.id

    @GObject.Property(type=Gtk.SortListModel, default=None)
    def model(self) -> Gtk.SortListModel:
        if self._model is None:
            filter_model = Gtk.FilterListModel.new(
                self._coremodel.props.songs)
            filter_model.set_filter(Gtk.AnyFilter())

            song_exp = Gtk.PropertyExpression.new(
                CoreSong, None, "track-number")
            song_sorter = Gtk.NumericSorter.new(song_exp)
            self._model = Gtk.SortListModel.new(filter_model, song_sorter)
            if self._model:
                self._model.connect("items-changed", self._on_disc_changed)

            self._coregrilo.get_album_disc(self, filter_model)

        return self._model

    def _on_disc_changed(self, model, position, removed, added):
        with self.freeze_notify():
            duration = 0
            for coresong in model:
                duration += coresong.props.duration

            self.props.duration = duration

    def remove_song_from_disc(self, song_urn: str) -> None:
        """Update this disc

        :param str song_urn: Song identifier
        """
        # FIXME: For now we just retrieve the full disc again
        filter_model = self.props.model.get_model()
        self._coregrilo.get_album_disc(self, filter_model)



# Copyright © 2018 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
import time
import typing

import gi
gi.require_version('GstPbutils', '1.0')
from gi.repository import GLib, GObject

from gnomemusic.coresong import CoreSong
from gnomemusic.gstplayer import GstPlayer, Playback
from gnomemusic.utils import RepeatMode
from gnomemusic.queue import Queue

if typing.TYPE_CHECKING:
    from gi.repository import Gio, Gtk


class Player(GObject.GObject):
    """Main Player object

    Contains the logic of playing a song with Music.
    """

    __gsignals__ = {
        'seek-finished': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'song-changed': (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    duration = GObject.Property(type=float, default=-1.)
    mute = GObject.Property(type=bool, default=False)
    state = GObject.Property(type=int, default=Playback.STOPPED)
    volume = GObject.Property(type=float, default=1.)

    def __init__(self, application):
        """Initialize the player

        :param Application application: Application object
        """
        super().__init__()

        self._app = application
        # In the case of gapless playback, both 'about-to-finish'
        # and 'eos' can occur during the same stream. 'about-to-finish'
        # already sets self._queue to the next song, so doing it
        # again on eos would skip a song.
        # TODO: Improve queue handling so this hack is no longer
        # needed.
        self._gapless_set = False
        self._log = application.props.log
        self._queue = Queue(self._app)

        self._queue_model = self._app.props.coremodel.props.queue
        self._queue_model.connect(
            "items-changed", self._on_queue_model_items_changed)

        self._settings = application.props.settings
        self._settings.connect("changed::repeat", self._on_repeat_mode_changed)

        self._repeat = RepeatMode(self._settings.get_enum("repeat"))
        self.bind_property(
            'repeat-mode', self._queue, 'repeat-mode',
            GObject.BindingFlags.SYNC_CREATE)

        self._new_clock = True

        self._gst_player = GstPlayer(application)
        self._gst_player.connect("about-to-finish", self._on_about_to_finish)
        self._gst_player.connect('clock-tick', self._on_clock_tick)
        self._gst_player.connect('eos', self._on_eos)
        self._gst_player.connect("error", self._on_error)
        self._gst_player.connect('seek-finished', self._on_seek_finished)
        self._gst_player.connect("stream-start", self._on_stream_start)
        self._gst_player.bind_property(
            'duration', self, 'duration', GObject.BindingFlags.SYNC_CREATE)
        self._gst_player.bind_property(
            'state', self, 'state', GObject.BindingFlags.SYNC_CREATE)

        self._gst_player.bind_property(
            "volume", self, "volume", GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

        self._gst_player.bind_property(
            "mute", self, "mute", GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

    @GObject.Property(
        type=bool, default=False, flags=GObject.ParamFlags.READABLE)
    def has_next(self):
        """Test if the queue has a next song.

        :returns: True if the current song is not the last one.
        :rtype: bool
        """
        return self._queue.has_next()

    @GObject.Property(
        type=bool, default=False, flags=GObject.ParamFlags.READABLE)
    def has_previous(self):
        """Test if the queue has a previous song.

        :returns: True if the current song is not the first one.
        :rtype: bool
        """
        return self._queue.has_previous()

    @GObject.Property(
        type=bool, default=False, flags=GObject.ParamFlags.READABLE)
    def playing(self):
        """Test if a song is currently played.

        :returns: True if a song is currently played.
        :rtype: bool
        """
        return self.props.state == Playback.PLAYING

    def _on_queue_model_items_changed(
            self, model: Gtk.SortListModel, position: int, removed: int,
            added: int) -> None:
        if (removed > 0
                and model.get_n_items() == 0):
            self.stop()

    def _on_repeat_mode_changed(
            self, settings: Gio.Settings, key: str) -> None:
        repeat_mode = settings.get_enum(key)
        self.props.repeat_mode = RepeatMode(repeat_mode)

    def _on_about_to_finish(self, klass):
        if self.props.has_next:
            self._log.debug("Song is about to finish, loading the next one.")
            next_coresong = self._queue.get_next()
            new_url = next_coresong.props.url
            self._gst_player.props.url = new_url
            self._gapless_set = True

    def _on_eos(self, klass):
        self._queue.next()

        if self._gapless_set:
            # After 'eos' in the gapless case, the pipeline needs to be
            # hard reset.
            self._log.debug("Song finished, loading the next one.")
            self.stop()
            self.play(self.props.current_song)
        else:
            self._log.debug("End of the queue, stopping the player.")
            self.stop()

        self._gapless_set = False

    def _on_error(self, klass=None):
        self.stop()
        self._gapless_set = False

        current_song = self.props.current_song
        current_song.props.validation = CoreSong.Validation.FAILED
        if (self.has_next
                and self.props.repeat_mode != RepeatMode.SONG):
            self.next()

    def _on_stream_start(self, klass):
        if self._gapless_set:
            self._queue.next()

        self._gapless_set = False
        self._time_stamp = int(time.time())

        self.emit("song-changed")

    def _load(self, coresong):
        self._log.debug("Loading song {}".format(coresong.props.title))
        self._gst_player.props.state = Playback.LOADING
        self._time_stamp = int(time.time())
        self._gst_player.props.url = coresong.props.url

    def play(self, coresong=None):
        """Play a song.

        Start playing a song, a specific CoreSong if supplied and
        available or a song in the queue decided by the play mode.

        If a song is paused, a subsequent play call without a CoreSong
        supplied will continue playing the paused song.

        :param CoreSong coresong: The CoreSong to play or None.
        """
        if self.props.current_song is None:
            coresong = self._queue.set_song(coresong)

        if (coresong is not None
                and coresong.props.validation == CoreSong.Validation.FAILED
                and self.props.repeat_mode != RepeatMode.SONG):
            self._on_error()
            return

        if coresong is not None:
            self._load(coresong)

        if self.props.current_song is not None:
            self._gst_player.props.state = Playback.PLAYING

    def pause(self):
        """Pause"""
        self._gst_player.props.state = Playback.PAUSED

    def stop(self):
        """Stop"""
        self._gst_player.props.state = Playback.STOPPED
        self._app.props.window.set_player_visible(False)
        self._queue.end()

    def next(self):
        """"Play next song

        Play the next song of the queue, if any.
        """
        if self._gapless_set:
            self.set_position(0.0)
        elif self._queue.next():
            self.play(self._queue.props.current_song)

    def previous(self):
        """Play previous song

        Play the previous song of the queue, if any.
        """
        position = self._gst_player.props.position
        if self._gapless_set:
            self.stop()

        if (position < 5
                and self._queue.has_previous()):
            self._queue.previous()
            self._gapless_set = False
            self.play(self._queue.props.current_song)
        # This is a special case for a song that is very short and the
        # first song in the queue. It can trigger gapless, but
        # has_previous will return False.
        elif (position < 5
                and self._queue.props.position == 0):
            self.set_position(0.0)
            self._gapless_set = False
            self.play(self._queue.props.current_song)
        else:
            self.set_position(0.0)

    def play_pause(self):
        """Toggle play/pause state"""
        if self.props.state == Playback.PLAYING:
            self.pause()
        else:
            self.play()

    def _on_clock_tick(self, klass, tick):
        self._log.debug("Clock tick {}, player at {} seconds".format(
            tick, self._gst_player.props.position))

        if tick == 0:
            self._new_clock = True

        if self.props.duration == -1.:
            return

        position = self._gst_player.props.position
        if position > 0:
            percentage = tick / self.props.duration
            if (percentage > 0.5
                    and self._new_clock):
                self._new_clock = False
                GLib.idle_add(self._update_stats)

    def _update_stats(self) -> None:
        current_song = self._queue.props.current_song
        current_song.props.last_played = GLib.DateTime.new_now_utc()
        current_song.bump_play_count()

    @GObject.Property(type=object)
    def repeat_mode(self) -> RepeatMode:
        """Gets current repeat mode.

        :returns: current repeat mode
        :rtype: RepeatMode
        """
        return self._repeat

    @repeat_mode.setter  # type: ignore
    def repeat_mode(self, mode):
        if mode == self._repeat:
            return

        self._repeat = mode
        self._settings.set_enum("repeat", mode.value)

    @GObject.Property(type=int, default=0, flags=GObject.ParamFlags.READABLE)
    def position(self):
        """Gets current song index.

        :returns: position of the current song in the queue.
        :rtype: int
        """
        return self._queue.props.position

    @GObject.Property(
        type=CoreSong, default=None, flags=GObject.ParamFlags.READABLE)
    def current_song(self):
        """Get the current song.

        :returns: The song being played. None if there is no queue.
        :rtype: CoreSong
        """
        return self._queue.props.current_song

    def get_position(self):
        """Get player position.

        Player position in seconds.
        :returns: position
        :rtype: float
        """
        return self._gst_player.props.position

    # TODO: used by MPRIS
    def set_position(self, position_second):
        """Change GstPlayer position.

        If the position if negative, set it to zero.
        If the position if greater than song duration, do nothing
        :param float position_second: requested position in second
        """
        if position_second < 0.0:
            position_second = 0.0

        duration_second = self._gst_player.props.duration
        if position_second <= duration_second:
            self._gst_player.seek(position_second)

    def _on_seek_finished(self, klass):
        # FIXME: Just a proxy
        self.emit('seek-finished')



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from typing import Any, Dict
import typing

from gi.repository import Gtk, GObject

from gnomemusic.artistart import ArtistArt
from gnomemusic.corealbum import CoreAlbum
import gnomemusic.utils as utils
if typing.TYPE_CHECKING:
    from gnomemusic.application import Application


class CoreArtist(GObject.GObject):
    """Artist information object

    Contains all relevant information about an artist.
    """

    artist = GObject.Property(type=str)
    id = GObject.Property(type=str)

    def __init__(
            self, application: Application,
            cursor_dict: Dict[str, Any]) -> None:
        """Initiate the CoreArtist object

        :param Application application: The application object
        :param Dict[str, Any] cursor_dict: Dict with Tsparql keys
        """
        super().__init__()

        self._application = application
        self._coregrilo = application.props.coregrilo
        self._coremodel = application.props.coremodel
        self._model = None
        self._thumbnail = None

        self.update(cursor_dict)

    def update(self, cursor_dict: Dict[str, Any]) -> None:
        self.props.id = cursor_dict.get("id")
        self.props.artist = utils.get_artist_from_cursor_dict(cursor_dict)

    def _get_artist_album_model(self):
        albums_model_filter = Gtk.FilterListModel.new(
            self._coremodel.props.albums)
        albums_model_filter.set_filter(Gtk.AnyFilter())

        albums_sort_exp = Gtk.PropertyExpression.new(CoreAlbum, None, "year")
        albums_sorter = Gtk.StringSorter.new(albums_sort_exp)
        albums_model_sort = Gtk.SortListModel.new(
            albums_model_filter, albums_sorter)

        self._coregrilo.get_artist_albums(self, albums_model_filter)

        return albums_model_sort

    @GObject.Property(type=Gtk.SortListModel, default=None)
    def model(self):
        if self._model is None:
            self._model = self._get_artist_album_model()

        return self._model

    @GObject.Property(type=str, default=None)
    def thumbnail(self):
        """Artist art thumbnail retrieval

        :return: The artist art uri or "generic"
        :rtype: string
        """
        if self._thumbnail is None:
            self._thumbnail = "generic"
            ArtistArt(self._application, self)

        return self._thumbnail

    @thumbnail.setter  # type: ignore
    def thumbnail(self, value):
        """Artist art thumbnail setter

        :param string value: uri or "generic"
        """
        self._thumbnail = value



# Copyright 2020 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

import asyncio

import gi
gi.require_version("MediaArt", "2.0")
from gi.repository import GLib, GObject, Gio, MediaArt

from gnomemusic.asynciolimiter import StrictLimiter
from gnomemusic.embeddedart import EmbeddedArt


class AlbumArt(GObject.GObject):
    """AlbumArt retrieval object
    """

    _limiter = StrictLimiter(5 / 1)

    def __init__(self, application, corealbum):
        """Initialize AlbumArt

        :param Application application: The application object
        :param CoreAlbum corealbum: The corealbum to use
        """
        super().__init__()

        self._corealbum = corealbum
        self._coregrilo = application.props.coregrilo
        self._album = self._corealbum.props.title
        self._artist = self._corealbum.props.artist

        asyncio.create_task(self._in_cache())

    def _on_embedded_art_found(self, embeddedart, found):
        if found:
            asyncio.create_task(self._in_cache())
        else:
            self._coregrilo.get_album_art(self._corealbum)

    async def _in_cache(self) -> None:
        success, thumb_file = MediaArt.get_file(
            self._artist, self._album, "album")
        if not success:
            return

        try:
            result = await thumb_file.query_info_async(
                Gio.FILE_ATTRIBUTE_STANDARD_TYPE, Gio.FileQueryInfoFlags.NONE,
                GLib.PRIORITY_DEFAULT_IDLE)
        except GLib.Error:
            # This indicates that the file has not been created, so
            # there is no art in the MediaArt cache.
            result = False

        if result:
            self._corealbum.props.thumbnail = thumb_file.get_uri()
        else:
            embedded = EmbeddedArt()
            embedded.connect("art-found", self._on_embedded_art_found)
            await self._limiter.wait()
            embedded.query(self._corealbum, self._album)



# Copyright (c) 2013 Arnel A. Borja <kyoushuu@yahoo.com>
# Copyright (c) 2013 Vadim Rutkovsky <vrutkovs@redhat.com>
# Copyright (c) 2013 Lubosz Sarnecki <lubosz@gmail.com>
# Copyright (c) 2013 Guillaume Quintard <guillaume.quintard@gmail.com>
# Copyright (c) 2013 Felipe Borges <felipe10borges@gmail.com>
# Copyright (c) 2013 Eslam Mostafa <cseslam@gmail.com>
# Copyright (c) 2013 Shivani Poddar <shivani.poddar92@gmail.com>
# Copyright (c) 2013 Sai Suman Prayaga <suman.sai14@gmail.com>
# Copyright (c) 2013 Seif Lotfy <seif@lotfy.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

import asyncio
from gettext import gettext as _

from gi.events import GLibEventLoopPolicy
from gi.repository import Adw, GLib, GObject, Gdk, Gio, GstAudio, Gtk

from gnomemusic.about import show_about
from gnomemusic.coregrilo import CoreGrilo
from gnomemusic.coremodel import CoreModel
from gnomemusic.inhibitsuspend import InhibitSuspend
from gnomemusic.mpris import MPRIS
from gnomemusic.musiclogger import MusicLogger
from gnomemusic.notificationmanager import NotificationManager
from gnomemusic.pauseonsuspend import PauseOnSuspend
from gnomemusic.player import Player
from gnomemusic.player import RepeatMode
from gnomemusic.search import Search
from gnomemusic.widgets.preferencesdialog import PreferencesDialog
from gnomemusic.window import Window


class Application(Adw.Application):

    def __init__(self, application_id, version):
        super().__init__(
            application_id=application_id,
            flags=Gio.ApplicationFlags.FLAGS_NONE)
        self.props.resource_base_path = "/org/gnome/Music"
        GLib.set_application_name(_("Music"))
        GLib.set_prgname(application_id)
        GLib.setenv("PULSE_PROP_application.id", application_id, True)

        asyncio.set_event_loop_policy(GLibEventLoopPolicy())

        self._version = version
        self._window = None

        self._log = MusicLogger()
        self._search = Search()

        self._notificationmanager = NotificationManager(self)
        self._coregrilo: CoreGrilo | None = None
        self._coremodel = CoreModel(self)

        self._settings = Gio.Settings.new('org.gnome.Music')
        self._player = Player(self)

        InhibitSuspend(self)

    @GObject.Property(
        type=CoreGrilo, default=None, flags=GObject.ParamFlags.READABLE)
    def coregrilo(self):
        """Get application-wide CoreGrilo instance.

        :returns: The grilo wrapper
        :rtype: CoreGrilo or None
        """
        return self._coregrilo

    @GObject.Property(
        type=MusicLogger, default=None, flags=GObject.ParamFlags.READABLE)
    def log(self):
        """Get application-wide logging facility.

        :returns: the logger
        :rtype: MusicLogger
        """
        return self._log

    @GObject.Property(
        type=Player, default=None, flags=GObject.ParamFlags.READABLE)
    def player(self):
        """Get application-wide music player.

        :returns: the player
        :rtype: Player
        """
        return self._player

    @GObject.Property(
        type=Gio.Settings, flags=GObject.ParamFlags.READABLE)
    def settings(self):
        """Get application-wide settings.

        :returns: settings
        :rtype: Gio.settings
        """
        return self._settings

    @GObject.Property(
        type=CoreModel, flags=GObject.ParamFlags.READABLE)
    def coremodel(self):
        """Get class providing all listmodels.

        :returns: List model provider class
        :rtype: CoreModel
        """
        return self._coremodel

    @GObject.Property(type=Window, flags=GObject.ParamFlags.READABLE)
    def window(self):
        """Get main window.

        :returns: Main window.
        :rtype: Window
        """
        return self._window

    @GObject.Property(
        type=Search, flags=GObject.ParamFlags.READABLE)
    def search(self):
        """Get class providing search logic.

        :returns: List model provider class
        :rtype: Search
        """
        return self._search

    @GObject.Property(
        type=NotificationManager, flags=GObject.ParamFlags.READABLE)
    def notificationmanager(self):
        """Get notification manager

        :returns: notification manager
        :rtype: NotificationManager
        """
        return self._notificationmanager

    def _set_actions(self) -> None:
        action_entries = [
            ("about", self._about, None),
            ("help", self._help, ("app.help", ["F1"])),
            ("mute", self._mute, ("app.mute", ["<Ctrl>M"])),
            ("play_pause", self._play_pause,
                ("app.play_pause",
                    ["<Ctrl>space", "AudioPlay", "AudioPause"])),
            ("preferences", self._preferences_dialog,
                ("app.preferences", ["<Ctrl>comma"])),
            ("quit", self._quit, ("app.quit", ["<Ctrl>Q"])),
            ("repeat_toggle", self._repeat_toggle,
                ("app.repeat_toggle", ["<Ctrl>R"])),
            ("shuffle_toggle", self._shuffle_toggle,
                ("app.shuffle_toggle", ["<Ctrl>S"])),
            ("song_next", self._song_next,
                ("app.song_next", ["<Ctrl>N", "AudioNext"])),
            ("song_previous", self._song_previous,
                ("app.song_previous", ["<Ctrl>B", "AudioPrev"])),
            ("song_stop", self._song_stop,
                ("app.song_stop", ["AudioStop"])),
            ("volume_decrease", self._volume_decrease,
                ("app.volume_decrease", ["<Ctrl>minus", "<Ctrl>KP_Subtract"])),
            ("volume_increase", self._volume_increase,
                ("app.volume_increase",
                    ["<Ctrl>plus", "<Ctrl>equal", "<Ctrl>KP_Add"])),
        ]

        for action, callback, accel in action_entries:
            simple_action = Gio.SimpleAction.new(action, None)
            simple_action.connect('activate', callback)
            self.add_action(simple_action)
            if accel:
                self.set_accels_for_action(*accel)

    def _about(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        show_about(self.props.application_id, self._version, self._window)

    def _help(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:

        def show_uri_cb(parent: Gtk.Window, result: Gio.AsyncResult) -> None:
            try:
                Gtk.show_uri_full_finish(parent, result)
            except GLib.Error:
                self._log.message("Help handler not available.")

        Gtk.show_uri_full(
            self._window, "help:gnome-music", Gdk.CURRENT_TIME, None,
            show_uri_cb)

    def _mute(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._player.props.mute = not self._player.props.mute

    def _play_pause(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._player.play_pause()

    def _preferences_dialog(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        if self._window.props.visible_dialog:
            return

        pref_dialog = PreferencesDialog(self)
        pref_dialog.present(self._window)

    def _quit(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._window.destroy()

    def _repeat_toggle(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        if self._player.props.repeat_mode == RepeatMode.SONG:
            self._player.props.repeat_mode = RepeatMode.ALL
        elif self._player.props.repeat_mode == RepeatMode.ALL:
            self._player.props.repeat_mode = RepeatMode.NONE
        else:
            self._player.props.repeat_mode = RepeatMode.SONG

    def _shuffle_toggle(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        if self._player.props.repeat_mode == RepeatMode.SHUFFLE:
            self._player.props.repeat_mode = RepeatMode.NONE
        else:
            self._player.props.repeat_mode = RepeatMode.SHUFFLE

    def _song_next(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._player.next()

    def _song_previous(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._player.previous()

    def _song_stop(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._player.stop()

    def _volume_decrease(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        if not self._player.props.mute:
            cubic_volume = GstAudio.stream_volume_convert_volume(
                GstAudio.StreamVolumeFormat.LINEAR,
                GstAudio.StreamVolumeFormat.CUBIC,
                self._player.props.volume)

            self._player.props.volume = GstAudio.stream_volume_convert_volume(
                GstAudio.StreamVolumeFormat.CUBIC,
                GstAudio.StreamVolumeFormat.LINEAR,
                max(0., cubic_volume - 0.1))

    def _volume_increase(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        if (not self._player.props.mute
                or self._player.props.volume == 0):
            cubic_volume = GstAudio.stream_volume_convert_volume(
                GstAudio.StreamVolumeFormat.LINEAR,
                GstAudio.StreamVolumeFormat.CUBIC,
                self._player.props.volume)

            self._player.props.volume = GstAudio.stream_volume_convert_volume(
                GstAudio.StreamVolumeFormat.CUBIC,
                GstAudio.StreamVolumeFormat.LINEAR,
                min(1., cubic_volume + 0.1))

        if self._player.props.mute:
            self._player.props.mute = False

    def do_startup(self):
        Adw.Application.do_startup(self)
        Adw.StyleManager.get_default().set_color_scheme(
            Adw.ColorScheme.PREFER_LIGHT)
        self._set_actions()

    def do_activate(self):
        self._coregrilo = CoreGrilo(self)

        if not self._window:
            self._window = Window(self)
            self.notify("window")
            self._window.set_default_icon_name(self.props.application_id)
            if self.props.application_id == "org.gnome.Music.Devel":
                self._window.get_style_context().add_class('devel')
            MPRIS(self)

        PauseOnSuspend(self._player)

        self._window.present()



# Copyright 2019 The GNOME Music developers
#
# GNOME Music is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# GNOME Music is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with GNOME Music; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# The GNOME Music authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and GNOME Music.  This permission is above and beyond the permissions
# granted by the GPL license by which GNOME Music is covered.  If you
# modify this code, you may extend this exception to your version of the
# code, but you are not obligated to do so.  If you do not wish to do so,
# delete this exception statement from your version.

from __future__ import annotations
from typing import Optional
import typing
import weakref

import gi
gi.require_version("Grl", "0.3")
from gi.repository import GLib, GObject, Gio, Grl, Gtk

from gnomemusic.grilowrappers.localsearchwrapper import LocalSearchWrapper
from gnomemusic.storeart import StoreArt
from gnomemusic.trackerwrapper import TrackerState, TrackerWrapper
from gnomemusic.utils import CoreObjectType
if typing.TYPE_CHECKING:
    from gnomemusic.corealbum import CoreAlbum
    from gnomemusic.coreartist import CoreArtist
    from gnomemusic.coredisc import CoreDisc
    from gnomemusic.coresong import CoreSong


class CoreGrilo(GObject.GObject):

    _METADATA_THUMBNAIL_KEYS = [
        Grl.METADATA_KEY_ID,
        Grl.METADATA_KEY_THUMBNAIL
    ]

    _blocklist = [
        'grl-bookmarks',
        'grl-filesystem',
        'grl-itunes-podcast',
        "grl-local-metadata",
        'grl-metadata-store',
        'grl-podcasts'
    ]

    _grl_plugin_ranks = ("grl-musicbrainz-coverart:3,"
                         "grl-lastfm-cover:2,"
                         "grl-theaudiodb-cover:1")

    _theaudiodb_api_key = "195003"

    cover_sources = GObject.Property(type=bool, default=False)
    tracker_available = GObject.Property(type=int)

    def __init__(self, application):
        """Initiate the CoreGrilo object

        :param Application application: The Application instance to use
        """
        super().__init__()

        self._application = application
        self._coremodel = self._application.props.coremodel
        self._log = application.props.log
        self._thumbnail_sources = []
        self._thumbnail_sources_timeout = None
        self._wrappers = {}

        self._fast_options: Grl.OperationOptions = Grl.OperationOptions()
        self._fast_options.set_resolution_flags(
            Grl.ResolutionFlags.FAST_ONLY | Grl.ResolutionFlags.IDLE_RELAY)

        self._tsparql_wrapper = TrackerWrapper(application)
        self._tsparql_wrapper.bind_property(
            "tracker-available", self, "tracker-available",
            GObject.BindingFlags.SYNC_CREATE)

        GLib.setenv("GRL_PLUGIN_RANKS", self._grl_plugin_ranks, True)

        Grl.init(None)

        self._registry = Grl.Registry.get_default()
        config = Grl.Config.new("grl-lua-factory", "grl-theaudiodb-cover")
        config.set_api_key(self._theaudiodb_api_key)
        self._registry.add_config(config)

        self._registry.connect('source-added', self._on_source_added)
        self._registry.connect('source-removed', self._on_source_removed)

        self._registry.load_all_plugins(False)

        tracker_available_state = self._tsparql_wrapper.props.tracker_available
        if tracker_available_state != TrackerState.AVAILABLE:
            self._tsparql_wrapper.connect(
                "notify::tracker-available",
                self._on_tracker_available_changed)
        else:
            self._on_tracker_available_changed(None, None)

        for plugin in self._registry.get_plugins(False):
            plugin_id = plugin.get_id()
            # Activate the Tracker plugin only when TrackerWrapper
            # is available by listening to the tracker-available
            # property, so skip it here.
            if plugin_id != "grl-tracker3":
                try:
                    self._registry.activate_plugin_by_id(plugin_id)
                except GLib.GError:
                    self._log.debug(
                        "Failed to activate {} plugin.".format(plugin_id))

        weakref.finalize(self, Grl.deinit)

    def _on_tracker_available_changed(
            self, trackerwrapper: TrackerWrapper, state: TrackerState) -> None:
        # FIXME:No removal support yet.
        new_state = self._tsparql_wrapper.props.tracker_available
        if new_state == TrackerState.AVAILABLE:
            wrapper = LocalSearchWrapper(
                self._application, self._tsparql_wrapper)
            self._wrappers["gnome-music"] = wrapper

    def _on_source_added(self, registry, source):

        def _trigger_art_update():
            self._thumbnail_sources_timeout = None
            if len(self._thumbnail_sources) > 0:
                self.props.cover_sources = True

            return GLib.SOURCE_REMOVE

        if ("net:plaintext" in source.get_tags()
                or source.props.source_id in self._blocklist):
            try:
                registry.unregister_source(source)
            except GLib.GError:
                self._log.warning(
                    "Failed to unregister {}".format(source.props.source_id))
            return

        if Grl.METADATA_KEY_THUMBNAIL in source.supported_keys():
            self._thumbnail_sources.append(source)
            if not self._thumbnail_sources_timeout:
                # Aggregate sources being added, for example when the
                # network comes online.
                self._thumbnail_sources_timeout = GLib.timeout_add_seconds(
                    5, _trigger_art_update)

    def _on_source_removed(self, registry, source):
        # FIXME: Handle removing sources.
        self._log.debug("Removed source {}".format(source.props.source_id))

    def get_artist_albums(
            self, coreartist: CoreArtist,
            filter_model: Gtk.FilterListModel) -> None:
        """Get all album by an artist

        :param CoreArtist coreart: An artist to look up
        :param Gtk.FilterListModel filter_model: The model to fill
        """
        source = "gnome-music"
        self._wrappers[source].get_artist_albums(coreartist, filter_model)

    def get_album_discs(
            self, corealbum: CoreAlbum, disc_model: Gio.ListStore) -> None:
        """Get all discs from an album

        :param CoreAlbum corealbum: An album
        :param Gtk.SortListModel disc_model: The model to fill
        """
        source = "gnome-music"
        self._wrappers[source].get_album_discs(corealbum, disc_model)

    def get_album_disc(
            self, coredisc: CoreDisc, model: Gtk.FilterListModel) -> None:
        """Get all songs from an album disc

        :param CoreDisc coredisc: An album disc to look up
        :param Gtk.FilterListModel model: The model to fill
        """
        source = "gnome-music"
        self._wrappers[source].get_album_disc(coredisc, model)

    def writeback_tracker(self, coresong: CoreSong, tag: str) -> None:
        """Use Tracker queries to update tags.

        The tags are associated with a Tracker resource
        (song, album, artist or external resource).

        :param CoreSong coresong: Song to update
        :param str tag: tag to update
        """
        self._tsparql_wrapper.update_tag(coresong, tag)

    def search(self, text: str) -> None:
        """Search for the given string in the wrappers

        If an empty string is provided, the wrapper should
        reset to an empty state.

        :param str text: The search string
        """
        for wrapper in self._wrappers.values():
            wrapper.search(text)

    def get_song_art(self, coresong: CoreSong) -> None:
        """Retrieve song art for the given CoreSong

        :param CoreSong coresong: CoreSong to retrieve art for
        """
        def _on_resolved(
                source: Grl.Source, op_id: int, media: Optional[Grl.Media],
                error: Optional[GLib.Error]) -> None:
            if error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            if media is None:
                return

            StoreArt().start(
                coresong, media.get_thumbnail(), CoreObjectType.SONG)

        for source in self._thumbnail_sources:
            media = Grl.Media.audio_new()
            media.set_album(coresong.props.album)
            media.set_artist(coresong.props.artist)
            media.set_url(coresong.props.url)
            source.resolve(
                media, self._METADATA_THUMBNAIL_KEYS, self._fast_options,
                _on_resolved)

    def get_album_art(self, corealbum: CoreAlbum) -> None:
        """Retrieve album art for the given CoreAlbum

        :param CoreAlbum corealbum: CoreAlbum to retrieve art for
        """
        def _on_resolved(
                source: Grl.Source, op_id: int, media: Optional[Grl.Media],
                error: Optional[GLib.Error]) -> None:
            if error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            if media is None:
                return

            StoreArt().start(
                corealbum, media.get_thumbnail(), CoreObjectType.ALBUM)

        for source in self._thumbnail_sources:
            # The grilo album field is used during resolve
            media = Grl.Media.audio_new()
            media.set_album(corealbum.props.title)
            media.set_artist(corealbum.props.artist)
            media.set_url(corealbum.props.url)
            source.resolve(
                media, self._METADATA_THUMBNAIL_KEYS, self._fast_options,
                _on_resolved)

    def get_artist_art(self, coreartist: CoreArtist) -> None:
        """Retrieve artist art for the given CoreArtist

        :param CoreArtist coreartist: CoreArtist to retrieve art for
        """
        def _on_resolved(
                source: Grl.Source, op_id: int, media: Optional[Grl.Media],
                error: Optional[GLib.Error]) -> None:
            if error:
                self._log.warning(f"Error: {error.domain}, {error.message}")
                return

            if media is None:
                return

            StoreArt().start(
                coreartist, media.get_thumbnail(), CoreObjectType.ARTIST)

        for source in self._thumbnail_sources:
            media = Grl.Media.audio_new()
            media.set_artist(coreartist.props.artist)
            source.resolve(
                media, self._METADATA_THUMBNAIL_KEYS, self._fast_options,
                _on_resolved)

    def stage_playlist_deletion(self, playlist):
        """Prepares playlist deletion.

        :param Playlist playlist: playlist
        """
        if "gnome-music" in self._wrappers:
            self._wrappers["gnome-music"].stage_playlist_deletion(
                playlist)

    def finish_playlist_deletion(self, playlist, deleted):
        """Finishes playlist deletion.

        :param Playlist playlist: playlist
        :param bool deleted: indicates if the playlist has been deleted
        """
        if "gnome-music" in self._wrappers:
            self._wrappers["gnome-music"].finish_playlist_deletion(
                playlist, deleted)

    def create_playlist(self, playlist_title, callback):
        """Creates a new user playlist.

        :param str playlist_title: playlist title
        :param callback: function to perform once, the playlist is created
        """
        if "gnome-music" in self._wrappers:
            self._wrappers["gnome-music"].create_playlist(
                playlist_title, callback)



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from gi.repository import Adw, GLib, GObject, Gdk, Gio, Gtk
from gettext import gettext as _

from gnomemusic.trackerwrapper import TrackerState
from gnomemusic.utils import View
from gnomemusic.views.albumsview import AlbumsView
from gnomemusic.views.artistsview import ArtistsView
from gnomemusic.views.searchview import SearchView
from gnomemusic.views.playlistsview import PlaylistsView
from gnomemusic.widgets.statusnavigationpage import StatusNavigationPage
from gnomemusic.widgets.headerbar import HeaderBar
from gnomemusic.widgets.playertoolbar import PlayerToolbar  # noqa: F401
from gnomemusic.widgets.playlistdialog import PlaylistDialog
from gnomemusic.windowplacement import WindowPlacement


@Gtk.Template(resource_path="/org/gnome/Music/ui/Window.ui")
class Window(Adw.ApplicationWindow):

    __gtype_name__ = "Window"

    active_view = GObject.Property(type=GObject.GObject, default=None)

    _loading_progress = Gtk.Template.Child()
    _main_navigation_page = Gtk.Template.Child()
    _main_toolbar_view = Gtk.Template.Child()
    _navigation_view = Gtk.Template.Child()
    _overlay = Gtk.Template.Child()
    _player_toolbar = Gtk.Template.Child()
    _stack = Gtk.Template.Child()
    _toast_overlay = Gtk.Template.Child()

    def __init__(self, app):
        """Initialize the main window.

        :param Gtk.Application app: Application object
        """
        super().__init__(application=app, title=_("Music"))

        self._app = app

        self.set_size_request(360, 294)
        WindowPlacement(self)

        self._headerbar = HeaderBar(self._app)
        self._search_view: SearchView | None = None
        self._status_navpage = StatusNavigationPage(app)
        self._playlist_dialog: PlaylistDialog | None = None

        self._player = app.props.player
        self._search = app.props.search

        self._shortcut_controller = Gtk.ShortcutController.new()
        self.add_controller(self._shortcut_controller)
        self._startup_timeout_id = 0
        self._setup_view()
        self._set_actions()

    def _setup_view(self):
        self._headerbar.props.stack = self._stack
        self._main_toolbar_view.add_top_bar(self._headerbar)

        self._search.bind_property(
            "search-mode-active", self._headerbar, "search-mode-active",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

        self._search.connect(
            "notify::search-mode-active", self._on_search_mode_changed)

        self._player_toolbar.props.player = self._player

        self.views = [None] * len(View)

        self._navigation_view.add(self._status_navpage)

        self._stack.connect(
            "notify::visible-child", self._on_stack_visible_child_changed)

        self._headerbar.props.state = HeaderBar.State.MAIN

        self._app.props.coremodel.connect(
            "notify::songs-available", self._on_songs_available)

        self._app.props.coregrilo.connect(
            "notify::tracker-available", self._on_tracker_available)

        def notify() -> bool:
            self._startup_timeout_id = 0
            self._app.props.coremodel.notify("songs-available")

            return GLib.SOURCE_REMOVE

        self._startup_timeout_id = GLib.timeout_add(1000, notify)

    def _switch_to_empty_view(self) -> None:
        tracker_state = self._app.props.coregrilo.props.tracker_available
        statusnp = self._status_navpage
        if tracker_state == TrackerState.UNAVAILABLE:
            statusnp.props.state = StatusNavigationPage.State.NO_TRACKER
        elif tracker_state == TrackerState.OUTDATED:
            statusnp.props.state = StatusNavigationPage.State.TRACKER_OUTDATED
        else:
            statusnp.props.state = StatusNavigationPage.State.EMPTY

    def _on_search_mode_changed(self, search, value):
        if self._search.props.search_mode_active:
            self._navigation_view.replace_with_tags(["searchview"])
        else:
            self._navigation_view.replace_with_tags(["mainview"])

    def _on_songs_available(self, klass, value):
        if self._startup_timeout_id > 0:
            GLib.source_remove(self._startup_timeout_id)
            self._startup_timeout_id = 0

        if self._app.props.coremodel.props.songs_available:
            self._navigation_view.replace_with_tags(["mainview"])
            self._switch_to_player_view()
        else:
            self._navigation_view.replace_with_tags(["status"])
            self._switch_to_empty_view()

    def _on_stack_visible_child_changed(self, klass, value):
        self.props.active_view = self._stack.props.visible_child

    def _on_tracker_available(self, klass, value):
        new_state = self._app.props.coregrilo.props.tracker_available

        if new_state != TrackerState.AVAILABLE:
            self._switch_to_empty_view()

        self._app.props.coremodel.notify("songs-available")

    def _switch_to_player_view(self):
        # All views are created together, so if the album view is
        # already initialized, assume the rest are as well.
        if self.views[View.ALBUM] is not None:
            return

        self._headerbar.props.state = HeaderBar.State.MAIN
        self.views[View.ALBUM] = AlbumsView(self._app)
        self.views[View.ARTIST] = ArtistsView(self._app)
        self.views[View.PLAYLIST] = PlaylistsView(self._app)

        self._search_view = SearchView(self._app)
        self._search.bind_property(
            "state", self._search_view, "search-state",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)
        self._navigation_view.add(self._search_view)
        self._search.bind_property(
            "search-mode-active", self._search_view, "search-mode-active",
            GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.SYNC_CREATE)

        for i in self.views[View.ALBUM:]:
            if i.props.title:
                stackpage = self._stack.add_titled(
                    i, i.props.name, i.props.title)
                stackpage.props.icon_name = i.props.icon_name
            else:
                self._stack.add_named(i, i.props.name)

    @GObject.Property(
        type=HeaderBar, default=None, flags=GObject.ParamFlags.READABLE)
    def headerbar(self) -> HeaderBar:
        """Get headerbar instance.

        :returns: The headerbar
        :rtype: HeaderBar
        """
        return self._headerbar

    @GObject.Property(
        type=Adw.NavigationView, default=None,
        flags=GObject.ParamFlags.READABLE)
    def navigation_view(self) -> Adw.NavigationView:
        """Get NavigationView instance.

        :returns: The navigation view
        :rtype: Adw.NavigationView
        """
        return self._navigation_view

    @Gtk.Template.Callback()
    def _on_key_press(
            self, controller: Gtk.EventControllerKey, keyval: int,
            keycode: int, state: Gdk.ModifierType) -> bool:
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        shift_mask = Gdk.ModifierType.SHIFT_MASK

        search_active = self._search.props.search_mode_active
        rename_active = getattr(
            self.views[View.PLAYLIST], "rename_active", False)
        unicode_char = chr(Gdk.keyval_to_unicode(keyval))

        # Open the search bar when typing printable chars.
        if ((not search_active
                and self._search_view is not None
                and self._is_main_view_active()
                and not keyval == Gdk.KEY_space)
                and GLib.unichar_isprint(unicode_char)
                and (modifiers == shift_mask
                     or modifiers == 0)
                and not rename_active):
            self._search.props.search_mode_active = True
            self._search_view.props.search_text = unicode_char
        else:
            return Gdk.EVENT_PROPAGATE

        return Gdk.EVENT_STOP

    def _set_actions(self) -> None:
        action_entries = [
            ("navigate_back", self._navigate_back, ["<Alt>Left"]),
            ("search_bar_close", self._search_bar_close, ["Escape"]),
            ("search_bar_open", self._search_bar_open, ["<Ctrl>F"]),
            ("view_albums", self._view_albums, ["<Alt>1", "<Alt>KP_1"]),
            ("view_artists", self._view_artists, ["<Alt>2", "<Alt>KP_2"]),
            ("view_playlists", self._view_playlists, ["<Alt>3", "<Alt>KP_3"])
        ]

        for action, callback, accel in action_entries:
            simple_action = Gio.SimpleAction.new(action, None)
            simple_action.connect("activate", callback)
            self.add_action(simple_action)
            if accel:
                shortcut = Gtk.Shortcut.new(
                    Gtk.ShortcutTrigger.parse_string("|".join(accel)),
                    Gtk.ShortcutAction.parse_string(f"action(win.{action})"))
                self._shortcut_controller.add_shortcut(shortcut)

    def _navigate_back(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        if not self._is_main_view_active():
            self._navigation_view.pop()

    def _search_bar_close(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        if self._search.props.search_mode_active:
            self._search.props.search_mode_active = False

    def _search_bar_open(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        rename_active = getattr(
            self.views[View.PLAYLIST], "rename_active", False)

        if not (rename_active
                or self._search.props.search_mode_active):
            self._search.props.search_mode_active = True

    def _view_albums(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._switch_to_view("albums")

    def _view_artists(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._switch_to_view("artists")

    def _view_playlists(
            self, action: Gio.SimpleAction,
            param: GLib.Variant | None) -> None:
        self._switch_to_view("playlists")

    def _switch_to_view(self, view_name: str) -> None:
        """Switch the view switcher to another page"""
        if self._is_main_view_active():
            self._stack.props.visible_child_name = view_name

    def _is_main_view_active(self) -> bool:
        """Returns True if the main (view switcher) navigation page
        is active
        """
        visible_page = self._navigation_view.props.visible_page
        return visible_page == self._main_navigation_page

    def set_player_visible(self, visible):
        """Set PlayWidget action visibility

        :param bool visible: actionbar visibility
        """
        self._player_toolbar.props.revealed = visible

    def loading_pulse(self) -> bool:
        """Pulse the loading progress bar

        :returns: GLib.SOURCE_CONTINUE
        :rtype: bool
        """
        self._loading_progress.pulse()

        return GLib.SOURCE_CONTINUE

    def loading_visible(self, show: bool) -> None:
        """Sets visibility of the loading progressbar

        :param bool show: Wheter to show the loading bar
        """
        self._loading_progress.props.visible = show



# Copyright 2025 The GNOME Music developers
#
# SPDX-License-Identifier: GPL-2.0-or-later WITH GStreamer-exception-2008

from __future__ import annotations
from random import sample
import typing

from gi.repository import Gio, GObject

from gnomemusic.musiclogger import MusicLogger

if typing.TYPE_CHECKING:
    from gnomemusic.coresong import CoreSong


class ShuffleListModel(GObject.GObject, Gio.ListModel):
    """Shuffles the underlying list model

    This works by keeping a local shuffle list and shuffling the
    values as needed.

    This is meant for the Queue and as such a position in the
    playlist is required for the shuffling. The model shuffles the
    items after the given position. When deshuffling it will only
    deshuffle the items that are still left in the queue.
    """

    __gtype_name__ = "ShuffleListModel"

    shuffled = GObject.Property(type=bool, default=False)

    def __init__(self, model: Gio.ListModel) -> None:
        """Initialize the list model

        :param Gio.ListModel model: Model to shuffle
        """
        super().__init__()

        self._log = MusicLogger()

        self._model = model
        self._model.connect("items-changed", self._on_items_changed)

        self._shuffle_values: list[int] = []

    def _on_items_changed(
            self, model: Gio.ListModel, position: int, removed: int,
            added: int) -> None:
        # FIXME: Deal with item changes during play
        n_items = model.get_n_items()
        self._shuffle_values = list(range(0, n_items))

    def do_get_item(self, position: int) -> CoreSong:
        return self._model.get_item(self._shuffle_values[position])

    def do_get_n_items(self):
        return self._model.get_n_items()

    def do_get_item_type(self):
        return self._model.get_item_type()

    def shuffle(
            self, position: int, initial_song_position: int = 0) -> None:
        """Shuffle the model

        :param int position: Shuffle the remaining items from this
            position on
        :param int initial_song_position: The song index where to
            start shuffling from
        """
        self.props.shuffled = True
        if self._model.get_n_items() == 0:
            return

        list_before = list(self._shuffle_values[:position + 1])
        values_after = self._shuffle_values[position + 1:]
        list_after = sample(values_after, len(values_after))
        self._shuffle_values = list_before + list_after

        self._shuffle_values.remove(initial_song_position)
        self._shuffle_values = [initial_song_position] + self._shuffle_values

        self._log.debug(f"Shuffled order: {self._shuffle_values}")

    def deshuffle(self, position: int | None = None) -> None:
        """Deshuffle the model

        :param int position: Deshuffle the remaining items from this
            position on
        """
        self.props.shuffled = False

        if position is not None:
            list_before = list(self._shuffle_values[:position])
            list_position = list(self._shuffle_values[position:position + 1])
            list_after = sorted(self._shuffle_values[position + 1:])
            self._shuffle_values = list_before + list_position + list_after
        else:
            self._shuffle_values = sorted(self._shuffle_values)

        self._log.debug(f"Deshuffled order: {self._shuffle_values}")
